# Copyright 2023 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing
from typing import Any, Dict, List, Tuple, Union

import pandas as pd
import scipy.special

import lale.docstrings
import lale.helpers
import lale.operators
from lale.expressions import count as agg_count
from lale.expressions import it, replace
from lale.expressions import sum as agg_sum
from lale.lib.category_encoders import target_encoder
from lale.lib.dataframe import get_columns

from ._util import get_obj_cols
from .aggregate import Aggregate
from .concat_features import ConcatFeatures
from .group_by import GroupBy
from .map import Map
from .monoid import Monoid, MonoidableOperator


class _TargetEncoderMonoid(Monoid):
    def __init__(
        self,
        *,
        feature_names_out: List[str],
        y_sum: float,
        y_count: int,
        col2cat2sum: Dict[str, Dict[Union[float, str], float]],
        col2cat2count: Dict[str, Dict[Union[float, str], int]],
    ):
        self.feature_names_out = feature_names_out
        assert set(col2cat2sum.keys()) == set(col2cat2count.keys())
        self.y_sum = y_sum
        self.y_count = y_count
        self.col2cat2sum = col2cat2sum
        self.col2cat2count = col2cat2count

    def combine(self, other: "_TargetEncoderMonoid"):
        assert list(self.feature_names_out) == list(other.feature_names_out)
        assert set(self.col2cat2sum.keys()) == set(other.col2cat2sum.keys())
        assert set(self.col2cat2count.keys()) == set(other.col2cat2count.keys())
        return _TargetEncoderMonoid(
            feature_names_out=self.feature_names_out,
            y_sum=self.y_sum + other.y_sum,
            y_count=self.y_count + other.y_count,
            col2cat2sum={
                col: {
                    cat: cat2sum.get(cat, 0) + other.col2cat2sum[col].get(cat, 0)
                    for cat in set(cat2sum.keys()) | set(other.col2cat2sum[col].keys())
                }
                for col, cat2sum in self.col2cat2sum.items()
            },
            col2cat2count={
                col: {
                    cat: cat2count.get(cat, 0) + other.col2cat2count[col].get(cat, 0)
                    for cat in set(cat2count.keys())
                    | set(other.col2cat2count[col].keys())
                }
                for col, cat2count in self.col2cat2count.items()
            },
        )


class _TargetEncoderImpl(MonoidableOperator[_TargetEncoderMonoid]):
    def __init__(self, **hyperparams):
        self._hyperparams = hyperparams

    def transform(self, X):
        if self._transformer is None:
            self._transformer = self._build_transformer()
        return self._transformer.transform(X)

    def get_feature_names_out(self):
        return self.feature_names_out_

    def from_monoid(self, monoid: _TargetEncoderMonoid):
        self._monoid = monoid
        self._transformer = None
        self.feature_names_out_ = monoid.feature_names_out
        self._prior = monoid.y_sum / monoid.y_count
        k = self._hyperparams["min_samples_leaf"]
        f = self._hyperparams["smoothing"]

        def blend(posterior, sample_size):
            if sample_size <= 1:
                return self._prior
            weighting = scipy.special.expit((sample_size - k) / f)
            return weighting * posterior + (1 - weighting) * self._prior

        self._col2cat2value = {
            col: {
                cat: blend(
                    monoid.col2cat2sum[col][cat] / monoid.col2cat2count[col][cat],
                    sample_size=monoid.col2cat2count[col][cat],
                )
                for cat in monoid.col2cat2sum[col].keys()
            }
            for col in monoid.col2cat2sum.keys()
        }

    def _build_transformer(self):
        categorical_features = self._hyperparams["cols"]

        def build_map_expr(col):
            if col not in categorical_features:
                return it[col]
            return replace(
                it[col],
                self._col2cat2value[col],
                handle_unknown="use_encoded_value",
                unknown_value=self._prior,
            )

        return Map(
            columns={col: build_map_expr(col) for col in self.feature_names_out_}
        )

    def to_monoid(self, batch: Tuple[Any, Any]):
        X, y = batch
        X_columns = list(typing.cast(List[str], get_columns(X)))
        y_name = lale.helpers.GenSym(set(X_columns))("target")
        if isinstance(y, pd.Series):
            y = pd.DataFrame({y_name: y})
        else:
            y = Map(columns={y_name: it[get_columns(y)[0]]}).transform(y)
        assert lale.helpers._is_df(y), type(y)
        classes = self._hyperparams["classes"]
        if classes is not None:
            ordinal_encoder = Map(
                columns={
                    y_name: replace(it[y_name], {v: i for i, v in enumerate(classes)})
                }
            )
            y = ordinal_encoder.transform(y)
        y_aggregator = Aggregate(
            columns={"sum": agg_sum(it[y_name]), "count": agg_count(it[y_name])}
        )
        y_aggregated = lale.helpers._ensure_pandas(y_aggregator.transform(y))
        y_sum = y_aggregated["sum"].iat[0]
        y_count = y_aggregated["count"].iat[0]
        Xy = ConcatFeatures.transform([X, y])
        if self._hyperparams["cols"] is None:
            self._hyperparams["cols"] = get_obj_cols(X)
        col2cat2sum: Dict[str, Dict[Union[float, str], float]] = {}
        col2cat2count: Dict[str, Dict[Union[float, str], int]] = {}
        for col in typing.cast(List[str], self._hyperparams["cols"]):
            pipeline = (
                Map(columns={col: it[col], y_name: it[y_name]})
                >> GroupBy(by=[it[col]])
                >> Aggregate(
                    columns={"sum": agg_sum(it[y_name]), "count": agg_count(it[y_name])}
                )
            )
            aggregated = lale.helpers._ensure_pandas(pipeline.transform(Xy))
            col2cat2sum[col] = aggregated["sum"].to_dict()
            col2cat2count[col] = aggregated["count"].to_dict()
        return _TargetEncoderMonoid(
            feature_names_out=X_columns,
            y_sum=y_sum,
            y_count=y_count,
            col2cat2sum=col2cat2sum,
            col2cat2count=col2cat2count,
        )


_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Relational algebra reimplementation of scikit-learn contrib's `TargetEncoder`_ transformer.
Works on both pandas and Spark dataframes by using `Aggregate`_ for `fit` and `Map`_ for `transform`, which in turn use the appropriate backend.

.. _`TargetEncoder`: https://contrib.scikit-learn.org/category_encoders/targetencoder.html
.. _`Aggregate`: https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.aggregate.html
.. _`Map`: https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.map.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.target_encoder.html",
    "type": "object",
    "tags": {
        "pre": ["categoricals"],
        "op": ["transformer", "interpretable"],
        "post": [],
    },
    "properties": {
        "hyperparams": target_encoder._hyperparams_schema,
        "input_fit": target_encoder._input_fit_schema,
        "input_transform": target_encoder._input_transform_schema,
        "output_transform": target_encoder._output_transform_schema,
    },
}

TargetEncoder = lale.operators.make_operator(_TargetEncoderImpl, _combined_schemas)

TargetEncoder = typing.cast(
    lale.operators.PlannedIndividualOp,
    TargetEncoder.customize_schema(
        classes={
            "anyOf": [
                {"enum": [None], "description": "Regression task."},
                {
                    "type": "array",
                    "items": {"type": "number"},
                    "description": "Classification task with numeric labels.",
                    "minItems": 2,
                },
                {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "Classification task with string labels.",
                    "minItems": 2,
                },
                {
                    "type": "array",
                    "items": {"type": "boolean"},
                    "description": "Classification task with Boolean labels.",
                    "minItems": 2,
                },
            ],
            "default": None,
        },
        drop_invariant={
            "enum": [False],
            "default": False,
            "description": "This implementation only supports `drop_invariant=False`.",
        },
        return_df={
            "enum": [True],
            "default": True,
            "description": "This implementation returns a pandas or spark dataframe if the input is a pandas or spark dataframe, respectively.",
        },
        handle_missing={
            "enum": ["value"],
            "default": "value",
            "description": "This implementation only supports `handle_missing='value'`.",
        },
        handle_unknown={
            "enum": ["value"],
            "default": "value",
            "description": "This implementation only supports `handle_unknown='value'`.",
        },
    ),
)

lale.docstrings.set_docstrings(TargetEncoder)



# Copyright 2020 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from typing import Any, Callable, Dict, Generic, List, Optional, TypeVar

import numpy as np
from typing_extensions import Protocol

from lale.helpers import _is_pandas_df, _is_spark_df

from ..dataframe import (
    column_index,
    count,
    get_columns,
    make_series_concat,
    make_series_distinct,
    select_col,
)
from .monoid import Monoid, MonoidFactory

try:
    import pyspark.sql.functions

    spark_installed = True
except ImportError:
    spark_installed = False


class _column_distinct_count_data(Monoid):
    def __init__(self, df, limit: Optional[int] = None):
        self.limit = limit
        self.df = make_series_distinct(df)

    def __len__(self):
        return count(self.df)

    @property
    def is_absorbing(self):
        if self.limit is None:
            return False
        else:
            return len(self) > self.limit

    def combine(self, other: "_column_distinct_count_data"):
        if self.is_absorbing:
            return self
        elif other.is_absorbing:
            return other
        else:
            c = make_series_concat(self.df, other.df)
            return _column_distinct_count_data(c, limit=self.limit)


# numpy or sparkdf or pandas
_Batch = Any


class count_distinct_column(MonoidFactory[_Batch, int, _column_distinct_count_data]):
    """
    Counts the number of distinct elements in a given column.  If a limit is specified,
    then, once the limit is reached, the count may no longer be accurate
    (but will always remain over the limit).
    """

    def __init__(self, col: column_index, limit: Optional[int] = None):
        self._col = col
        self._limit = limit

    def to_monoid(self, batch) -> _column_distinct_count_data:
        c = select_col(batch, self._col)
        return _column_distinct_count_data(c, limit=self._limit)

    def from_monoid(self, monoid: _column_distinct_count_data) -> int:
        return len(monoid)


class categorical_column(MonoidFactory[_Batch, bool, _column_distinct_count_data]):
    """
    Determines if a column should be considered categorical,
    by seeing if there are more than threshold distinct values in it
    """

    def __init__(self, col: column_index, threshold: int = 5):
        self._col = col
        self._threshold = threshold

    def to_monoid(self, batch) -> _column_distinct_count_data:
        c = select_col(batch, self._col)
        return _column_distinct_count_data(c, limit=self._threshold)

    def from_monoid(self, monoid: _column_distinct_count_data) -> bool:
        return not monoid.is_absorbing


class make_categorical_column:
    def __init__(self, threshold=5):
        self._threshold = threshold

    def __call__(self, col):
        return categorical_column(col, threshold=self._threshold)


_D = TypeVar("_D", bound=Monoid)


class DictMonoid(Generic[_D], Monoid):
    """
    Given a monoid, this class lifts it to a dictionary pointwise
    """

    def __init__(self, m: Dict[Any, _D]):
        self._m = m

    def combine(self, other: "DictMonoid[_D]"):
        r = {k: self._m[k].combine(other._m[k]) for k in self._m.keys()}
        return DictMonoid(r)

    @property
    def is_absorbing(self):
        return all(v.is_absorbing for v in self._m.values())


class ColumnSelector(MonoidFactory[_Batch, List[column_index], _D], Protocol):
    def __call__(self, df) -> List[column_index]:
        return self.from_monoid(self.to_monoid(df))


class ColumnMonoidFactory(ColumnSelector[DictMonoid[_D]]):
    """
    Given a MonoidFactory for deciding if a given column is valid,
    This returns the list of valid columns
    """

    _makers: Optional[Dict[column_index, MonoidFactory[_Batch, bool, _D]]]

    def __init__(
        self, col_maker: Callable[[column_index], MonoidFactory[_Batch, bool, _D]]
    ):
        self._col_maker = col_maker
        self._makers = None

    def _get_makers(self, df):
        makers = self._makers
        if makers is None:
            indices = get_columns(df)
            makers = {k: self._col_maker(k) for k in indices}
            self._makers = makers
        return makers

    def to_monoid(self, batch):
        makers = self._get_makers(batch)
        return DictMonoid({k: v.to_monoid(batch) for k, v in makers.items()})

    def from_monoid(self, monoid: DictMonoid[_D]) -> List[column_index]:
        makers = self._makers
        assert makers is not None
        return [k for k, v in makers.items() if v.from_monoid(monoid._m[k])]


class categorical(ColumnMonoidFactory):
    """Creates a MonoidFactory (and callable) for projecting categorical columns with sklearn's ColumnTransformer or Lale's Project operator.

    Parameters
    ----------
    max_values : int

        Maximum number of unique values in a column for it to be considered categorical.

    Returns
    -------
    callable
        Function that, given a dataset X, returns a list of columns,
        containing either string column names or integer column indices."""

    def __init__(self, max_values: int = 5):
        super().__init__(make_categorical_column(max_values))


class date_time:
    """Creates a callable for projecting date/time columns with sklearn's ColumnTransformer or Lale's Project operator.

    Parameters
    ----------
    fmt : str

        Format string for `strptime()`, see https://docs.python.org/3/library/datetime.html#strftime-strptime-behavior

    Returns
    -------
    callable
        Function that, given a dataset X, returns a list of columns,
        containing either string column names or integer column indices."""

    def __init__(self, fmt):
        self._fmt = fmt

    def __repr__(self):
        return f"lale.lib.rasl.date_time(fmt={self._fmt})"

    def __call__(self, X):
        def is_date_time(column_values):
            try:
                for val in column_values:
                    if isinstance(val, str):
                        datetime.datetime.strptime(val, self._fmt)
                    else:
                        return False
            except ValueError:
                return False
            return True

        if _is_pandas_df(X):
            result = [c for c in X.columns if is_date_time(X[c])]
        elif isinstance(X, np.ndarray):
            result = [c for c in range(X.shape[1]) if is_date_time(X[:, c])]
        else:
            raise TypeError(f"unexpected type {type(X)}")
        return result


# functions for filter
def filter_isnan(df: Any, column_name: str):
    if _is_pandas_df(df):
        return df[df[column_name].isnull()]
    elif spark_installed and _is_spark_df(df):
        return df.filter(pyspark.sql.functions.isnan(df[column_name]))
    else:
        raise ValueError(
            "the filter isnan supports only Pandas dataframes or spark dataframes."
        )


def filter_isnotnan(df: Any, column_name: str):
    if _is_pandas_df(df):
        return df[df[column_name].notnull()]
    elif spark_installed and _is_spark_df(df):
        return df.filter(~pyspark.sql.functions.isnan(df[column_name]))
    else:
        raise ValueError(
            "the filter isnotnan supports only Pandas dataframes or spark dataframes."
        )


def filter_isnull(df: Any, column_name: str):
    if _is_pandas_df(df):
        return df[df[column_name].isnull()]
    elif spark_installed and _is_spark_df(df):
        return df.filter(pyspark.sql.functions.isnull(df[column_name]))
    else:
        raise ValueError(
            "the filter isnan supports only Pandas dataframes or spark dataframes."
        )


def filter_isnotnull(df: Any, column_name: str):
    if _is_pandas_df(df):
        return df[df[column_name].notnull()]
    elif spark_installed and _is_spark_df(df):
        return df.filter(~pyspark.sql.functions.isnull(df[column_name]))
    else:
        raise ValueError(
            "the filter isnotnan supports only Pandas dataframes or spark dataframes."
        )



# Copyright 2019-2022 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import sklearn.linear_model

import lale.operators

from .mock_module import CustomOrigOperator as model_to_be_wrapped


class _IncreaseRowsImpl:
    def __init__(self, n_rows=5):
        self.n_rows = n_rows

    def transform(self, X, y=None):
        subset_X = X[0 : self.n_rows]
        output_X = np.concatenate((X, subset_X), axis=0)
        return output_X

    def transform_X_y(self, X, y):
        output_X = self.transform(X)
        if y is None:
            output_y = None
        else:
            subset_y = y[0 : self.n_rows]
            output_y = np.concatenate((y, subset_y), axis=0)
        return output_X, output_y


_input_transform_schema_ir = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        },
    },
}

_output_transform_schema_ir = {
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}

_input_transform_X_y_schema_ir = {
    "type": "object",
    "required": ["X", "y"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        },
        "y": {
            "anyOf": [
                {"enum": [None]},
                {"type": "array", "items": {"type": "number"}},
            ],
        },
    },
}

_output_transform_X_y_schema_ir = {
    "type": "array",
    "laleType": "tuple",
    "items": [
        {
            "description": "X",
            "type": "array",
            "items": {"type": "array", "items": {"type": "number"}},
        },
        {
            "description": "y",
            "anyOf": [
                {"enum": [None]},
                {"type": "array", "items": {"type": "number"}},
            ],
        },
    ],
}

_hyperparam_schema_ir = {
    "allOf": [
        {
            "type": "object",
            "additionalProperties": False,
            "relevantToOptimizer": [],
            "properties": {"n_rows": {"type": "integer", "minimum": 0, "default": 5}},
        }
    ],
}

_combined_schemas_ir = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Combined schema for expected data and hyperparameters.",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparam_schema_ir,
        "input_transform": _input_transform_schema_ir,
        "output_transform": _output_transform_schema_ir,
        "input_transform_X_y": _input_transform_X_y_schema_ir,
        "output_transform_X_y": _output_transform_X_y_schema_ir,
    },
}

IncreaseRows = lale.operators.make_operator(_IncreaseRowsImpl, _combined_schemas_ir)


class _MyLRImpl:
    _wrapped_model: sklearn.linear_model.LogisticRegression

    def __init__(self, penalty="l2", solver="liblinear", C=1.0):
        self.penalty = penalty
        self.solver = solver
        self.C = C

    def fit(self, X, y):
        result = _MyLRImpl(self.penalty, self.solver, self.C)
        result._wrapped_model = sklearn.linear_model.LogisticRegression(
            penalty=self.penalty, solver=self.solver, C=self.C
        )
        result._wrapped_model.fit(X, y)
        return result

    def predict(self, X):
        assert hasattr(self, "_wrapped_model")
        return self._wrapped_model.predict(X)


_input_fit_schema_lr = {
    "type": "object",
    "required": ["X", "y"],
    "additionalProperties": False,
    "properties": {
        "X": {"type": "array", "items": {"type": "array", "items": {"type": "number"}}},
        "y": {"type": "array", "items": {"type": "number"}},
    },
}

_input_predict_schema_lr = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {"type": "array", "items": {"type": "array", "items": {"type": "number"}}}
    },
}

_output_predict_schema_lr = {"type": "array", "items": {"type": "number"}}

_hyperparams_ranges = {
    "type": "object",
    "additionalProperties": False,
    "required": ["solver", "penalty", "C"],
    "relevantToOptimizer": ["solver", "penalty", "C"],
    "properties": {
        "solver": {
            "description": "Algorithm for optimization problem.",
            "enum": ["newton-cg", "lbfgs", "liblinear", "sag", "saga"],
            "default": "liblinear",
        },
        "penalty": {
            "description": "Norm used in the penalization.",
            "enum": ["l1", "l2"],
            "default": "l2",
        },
        "C": {
            "description": "Inverse regularization strength. Smaller values specify "
            "stronger regularization.",
            "type": "number",
            "distribution": "loguniform",
            "minimum": 0.0,
            "exclusiveMinimum": True,
            "default": 1.0,
            "minimumForOptimizer": 0.03125,
            "maximumForOptimizer": 32768,
        },
    },
}

_hyperparams_constraints = {
    "allOf": [
        {
            "description": "The newton-cg, sag, and lbfgs solvers support only l2 penalties.",
            "anyOf": [
                {
                    "type": "object",
                    "properties": {
                        "solver": {"not": {"enum": ["newton-cg", "sag", "lbfgs"]}}
                    },
                },
                {"type": "object", "properties": {"penalty": {"enum": ["l2"]}}},
            ],
        }
    ]
}

_hyperparams_schema_lr = {"allOf": [_hyperparams_ranges, _hyperparams_constraints]}

_combined_schemas_lr = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "classifier"], "post": []},
    "properties": {
        "input_fit": _input_fit_schema_lr,
        "input_predict": _input_predict_schema_lr,
        "output_predict": _output_predict_schema_lr,
        "hyperparams": _hyperparams_schema_lr,
    },
}

MyLR = lale.operators.make_operator(_MyLRImpl, _combined_schemas_lr)


class _CustomParamsCheckerOpImpl:
    def __init__(self, fit_params=None, predict_params=None):
        self._fit_params = fit_params
        self._predict_params = predict_params

    def fit(self, X, y=None, **kwargs):
        result = _CustomParamsCheckerOpImpl(kwargs, self._predict_params)
        return result

    def predict(self, X, **predict_params):
        self._predict_params = predict_params


_input_fit_schema_cp = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "type": "object",
    "required": ["X", "y"],
    "additionalProperties": False,
    "properties": {
        "X": {"type": "array", "items": {"type": "array", "items": {"type": "number"}}},
        "y": {"items": {"type": "array", "items": {"type": "number"}}},
    },
}

_input_predict_schema_cp = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Predict using the linear model",
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "number",
        }
    },
}

_output_predict_schema_cp = {
    "description": "Returns predicted values.",
    "type": "array",
    "items": {"type": "number"},
}

_hyperparam_schema_cp = {
    "allOf": [
        {
            "description": "This first sub-object lists all constructor arguments with their "
            "types, one at a time, omitting cross-argument constraints.",
            "type": "object",
            "additionalProperties": False,
            "relevantToOptimizer": [],
            "properties": {},
        }
    ],
}

_combined_schemas_cp = {
    "description": "Combined schema for expected data and hyperparameters.",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator"], "post": []},
    "properties": {
        "hyperparams": _hyperparam_schema_cp,
        "input_fit": _input_fit_schema_cp,
        "input_predict": _input_predict_schema_cp,
        "output_predict": _output_predict_schema_cp,
    },
}

CustomParamsCheckerOp = lale.operators.make_operator(
    _CustomParamsCheckerOpImpl, _combined_schemas_cp
)

CustomOrigOperator = lale.operators.make_operator(model_to_be_wrapped, {})


class _OpThatWorksWithFilesImpl:
    def __init__(self, ngram_range):
        self.ngram_range = ngram_range

    def fit(self, X, y=None, sample_weight=None):
        import pandas as pd

        assert (
            sample_weight is not None
        ), "This is to test that Hyperopt passes fit_params correctly."
        from lale.lib.sklearn import LogisticRegression, TfidfVectorizer

        self.pipeline = (
            TfidfVectorizer(input="content", ngram_range=self.ngram_range)
            >> LogisticRegression()
        )
        self._wrapped_model = self.pipeline.fit(pd.read_csv(X, header=None), y)
        return self

    def predict(self, X):
        assert hasattr(self, "_wrapped_model")
        import pandas as pd

        return self._wrapped_model.predict(pd.read_csv(X, header=None))


_hyperparams_ranges_OpThatWorksWithFilesImpl = {
    "type": "object",
    "additionalProperties": False,
    "required": ["ngram_range"],
    "relevantToOptimizer": ["ngram_range"],
    "properties": {
        "ngram_range": {
            "default": (1, 1),
            "anyOf": [
                {
                    "type": "array",
                    "laleType": "tuple",
                    "minItemsForOptimizer": 2,
                    "maxItemsForOptimizer": 2,
                    "items": {
                        "type": "integer",
                        "minimumForOptimizer": 1,
                        "maximumForOptimizer": 3,
                    },
                    "forOptimizer": False,
                },
                {"enum": [(1, 1), (1, 2), (1, 3), (2, 2), (2, 3), (3, 3)]},
            ],
        },
    },
}

_hyperparams_schema_OpThatWorksWithFilesImpl = {
    "allOf": [_hyperparams_ranges_OpThatWorksWithFilesImpl]
}

_combined_schemas_OpThatWorksWithFilesImpl = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "classifier"], "post": []},
    "properties": {
        "input_fit": {},
        "input_predict": {},
        "output_predict": {},
        "hyperparams": _hyperparams_schema_OpThatWorksWithFilesImpl,
    },
}

OpThatWorksWithFiles = lale.operators.make_operator(
    _OpThatWorksWithFilesImpl, _combined_schemas_OpThatWorksWithFilesImpl
)



# Copyright 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Tuple

import numpy as np

import lale.docstrings
import lale.operators
from lale.expressions import it
from lale.lib.dataframe import count, get_columns
from lale.lib.rasl import Map
from lale.lib.sklearn import select_k_best

from .monoid import Monoid, MonoidableOperator
from .scores import FClassif


class _SelectKBestMonoid(Monoid):
    def __init__(self, *, n_samples_seen_, feature_names_in_, lifted_score_):
        self.n_samples_seen_ = n_samples_seen_
        self.feature_names_in_ = feature_names_in_
        self.lifted_score_ = lifted_score_

    def combine(self, other: "_SelectKBestMonoid"):
        n_samples_seen_ = self.n_samples_seen_ + other.n_samples_seen_
        assert list(self.feature_names_in_) == list(other.feature_names_in_)
        feature_names_in_ = self.feature_names_in_
        lifted_score_ = self.lifted_score_.combine(other.lifted_score_)
        return _SelectKBestMonoid(
            n_samples_seen_=n_samples_seen_,
            feature_names_in_=feature_names_in_,
            lifted_score_=lifted_score_,
        )


class _SelectKBestImpl(MonoidableOperator[_SelectKBestMonoid]):
    def __init__(self, monoidable_score_func=FClassif, score_func=None, *, k=10):
        self._hyperparams = {
            "score_func": monoidable_score_func(),
            "k": k,
        }

    def transform(self, X):
        if self._transformer is None:
            self._transformer = self._build_transformer()
        return self._transformer.transform(X)

    @property
    def n_samples_seen_(self):
        return getattr(self._monoid, "n_samples_seen_", 0)

    @property
    def feature_names_in_(self):
        return getattr(self._monoid, "feature_names_in_", None)

    def from_monoid(self, monoid: _SelectKBestMonoid):
        self._monoid = monoid
        score_func = self._hyperparams["score_func"]
        lifted_score_ = self._monoid.lifted_score_
        self.scores_, self.pvalues_ = score_func.from_monoid(lifted_score_)
        self.n_features_in_ = len(self._monoid.feature_names_in_)
        self._transformer = None

    def _build_transformer(self):
        assert self._monoid is not None
        k = self._hyperparams["k"]
        scores = self.scores_.copy()
        scores[np.isnan(scores)] = np.finfo(scores.dtype).min
        ind = np.sort(np.argpartition(scores, -min(k, len(scores)))[-k:])
        kbest = self._monoid.feature_names_in_[ind]
        result = Map(columns={col: it[col] for col in kbest})
        return result

    def to_monoid(self, batch: Tuple[Any, Any]):
        X, y = batch
        score_func = self._hyperparams["score_func"]
        n_samples_seen_ = count(X)
        feature_names_in_ = get_columns(X)
        lifted_score_ = score_func.to_monoid((X, y))
        return _SelectKBestMonoid(
            n_samples_seen_=n_samples_seen_,
            feature_names_in_=feature_names_in_,
            lifted_score_=lifted_score_,
        )


_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Relational algebra implementation of SelectKBest.",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.select_k_best.html",
    "type": "object",
    "tags": {
        "pre": ["~categoricals"],
        "op": ["transformer", "interpretable"],
        "post": [],
    },
    "properties": {
        "hyperparams": select_k_best._hyperparams_schema,
        "input_fit": select_k_best._input_fit_schema,
        "input_transform": select_k_best._input_transform_schema,
        "output_transform": select_k_best._output_transform_schema,
    },
}

SelectKBest: lale.operators.PlannedIndividualOp
SelectKBest = lale.operators.make_operator(_SelectKBestImpl, _combined_schemas)

lale.docstrings.set_docstrings(SelectKBest)



# Copyright 2021-2023 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from test.test_relational_sklearn import (
    _check_trained_min_max_scaler,
    _check_trained_one_hot_encoder,
    _check_trained_ordinal_encoder,
    _check_trained_select_k_best,
    _check_trained_simple_imputer,
    _check_trained_standard_scaler,
    _check_trained_target_encoder,
)

import jsonschema
import numpy as np
import pandas as pd
import sklearn
from category_encoders import TargetEncoder as SkTargetEncoder
from sklearn.datasets import fetch_openml, load_digits
from sklearn.feature_selection import SelectKBest as SkSelectKBest
from sklearn.impute import SimpleImputer as SkSimpleImputer
from sklearn.preprocessing import MinMaxScaler as SkMinMaxScaler
from sklearn.preprocessing import OneHotEncoder as SkOneHotEncoder
from sklearn.preprocessing import OrdinalEncoder as SkOrdinalEncoder
from sklearn.preprocessing import StandardScaler as SkStandardScaler

from lale.helpers import _ensure_pandas
from lale.lib.rasl import Convert
from lale.lib.rasl import MinMaxScaler as RaslMinMaxScaler
from lale.lib.rasl import OneHotEncoder as RaslOneHotEncoder
from lale.lib.rasl import OrdinalEncoder as RaslOrdinalEncoder
from lale.lib.rasl import SelectKBest as RaslSelectKBest
from lale.lib.rasl import SimpleImputer as RaslSimpleImputer
from lale.lib.rasl import StandardScaler as RaslStandardScaler
from lale.lib.rasl import TargetEncoder as RaslTargetEncoder

assert sklearn.__version__ >= "1.0", sklearn.__version__


def _check_data(self, sk_data, rasl_data, msg):
    rasl_data = _ensure_pandas(rasl_data)
    self.assertEqual(sk_data.shape, rasl_data.shape, msg)
    for row_idx in range(sk_data.shape[0]):
        for col_idx in range(sk_data.shape[1]):
            if isinstance(sk_data, np.ndarray):
                sk_val = sk_data[row_idx, col_idx]
            else:
                sk_val = sk_data.iloc[row_idx, col_idx]
            rasl_val = rasl_data.iloc[row_idx, col_idx]
            if isinstance(sk_val, np.number) and np.isnan(sk_val):
                self.assertTrue(np.isnan(rasl_val))
            else:
                self.assertAlmostEqual(sk_val, rasl_val, msg=(row_idx, col_idx, msg))


class TestMinMaxScaler(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.targets = ["pandas", "spark"]

    def test1(self):
        """
        From https://scikit-learn.org/1.1/modules/generated/sklearn.preprocessing.MinMaxScaler.html
        >>> from sklearn.preprocessing import MinMaxScaler
        >>> data = [[-1, 2], [-0.5, 6], [0, 10], [1, 18]]
        >>> scaler = MinMaxScaler()
        >>> print(scaler.fit(data))
        MinMaxScaler()
        >>> print(scaler.data_max_)
        [ 1. 18.]
        >>> print(scaler.transform(data))
        [[0.   0.  ]
        [0.25 0.25]
        [0.5  0.5 ]
        [1.   1.  ]]
        >>> print(scaler.transform([[2, 2]]))
        [[1.5 0. ]]
        """
        data = [[-1, 2], [-0.5, 6], [0, 10], [1, 18]]
        sk_scaler = SkMinMaxScaler()
        sk_scaler.fit(data)
        sk_transformed_data = sk_scaler.transform(data)
        X = [[2, 2]]
        sk_transformed_X = sk_scaler.transform(X)

        for target in self.targets:
            data = Convert(astype=target).transform(data)
            rasl_scaler = RaslMinMaxScaler()
            rasl_scaler.fit(data)
            _check_trained_min_max_scaler(self, sk_scaler, rasl_scaler, target)
            rasl_transformed_data = rasl_scaler.transform(data)
            _check_data(self, sk_transformed_data, rasl_transformed_data, target)
            X = Convert(astype=target).transform(X)
            rasl_transformed_X = rasl_scaler.transform(X)
            _check_data(self, sk_transformed_X, rasl_transformed_X, target)


class TestOrdinalEncoder(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.targets = ["pandas", "spark"]

    def test_1(self):
        """
        From https://scikit-learn.org/1.1/modules/generated/sklearn.preprocessing.OrdinalEncoder.html
        >>> from sklearn.preprocessing import OrdinalEncoder
        >>> enc = OrdinalEncoder()
        >>> X = [['Male', 1], ['Female', 3], ['Female', 2]]
        >>> enc.fit(X)
        OrdinalEncoder()
        >>> enc.categories_
        [array(['Female', 'Male'], dtype=object), array([1, 2, 3], dtype=object)]
        >>> enc.transform([['Female', 3], ['Male', 1]])
        array([[0., 2.],
            [1., 0.]])
        """
        sk_enc = SkOrdinalEncoder()
        X = [["Male", 1], ["Female", 3], ["Female", 2]]
        sk_enc.fit(X)
        data = [["Female", 3], ["Male", 1]]
        sk_transformed = sk_enc.transform(data)
        for target in self.targets:
            rasl_enc = RaslOrdinalEncoder()
            X = Convert(astype=target).transform(X)
            rasl_enc.fit(X)
            data = Convert(astype=target).transform(data)
            rasl_transformed = rasl_enc.transform(data)
            _check_trained_ordinal_encoder(self, sk_enc, rasl_enc, target)
            _check_data(self, sk_transformed, rasl_transformed, target)

    def test_2(self):
        """
        From https://scikit-learn.org/1.1/modules/generated/sklearn.preprocessing.OrdinalEncoder.html
        >>> import numpy as np
        >>> X = [['Male', 1], ['Female', 3], ['Female', np.nan]]
        >>> enc.fit_transform(X)
        array([[ 1.,  0.],
            [ 0.,  1.],
            [ 0., nan]])
        """
        sk_enc = SkOrdinalEncoder()
        X = [["Male", 1], ["Female", 3], ["Female", np.nan]]
        sk_transformed = sk_enc.fit_transform(X)
        for target in self.targets:
            rasl_enc = RaslOrdinalEncoder()
            X = Convert(astype=target).transform(X)
            rasl_transformed = rasl_enc.fit_transform(X)
            _check_trained_ordinal_encoder(self, sk_enc, rasl_enc, target)
            if target == "spark":
                continue  # XXX issue with NaN
            _check_data(self, sk_transformed, rasl_transformed, target)

    def test_3(self):
        """
        From https://scikit-learn.org/1.1/modules/generated/sklearn.preprocessing.OrdinalEncoder.html
        >>> enc.set_params(encoded_missing_value=-1).fit_transform(X)
        array([[ 1.,  0.],
            [ 0.,  1.],
            [ 0., -1.]])
        """
        pass  # XXX encoded_missing_value is not implemented


class TestSelectKBest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.targets = [
            "pandas",
            "spark",
        ]

    def test1(self):
        """
        From https://scikit-learn.org/1.1/modules/generated/sklearn.feature_selection.SelectKBest.html
        >>> from sklearn.datasets import load_digits
        >>> from sklearn.feature_selection import SelectKBest, chi2
        >>> X, y = load_digits(return_X_y=True)
        >>> X.shape
        (1797, 64)
        >>> X_new = SelectKBest(chi2, k=20).fit_transform(X, y)
        >>> X_new.shape
        (1797, 20)
        """

        X, y = load_digits(return_X_y=True, as_frame=True)
        sk_selectkbest = SkSelectKBest(k=20)
        sk_X_new = sk_selectkbest.fit_transform(
            X, y
        )  # XXX chi2 is not supported in RASL
        for target in self.targets:
            X = Convert(astype=target).transform(X)
            y = Convert(astype=target).transform(y)
            rasl_selectkbest = RaslSelectKBest(k=20)
            rasl_X_new = rasl_selectkbest.fit_transform(X, y)
            _check_trained_select_k_best(self, sk_selectkbest, rasl_selectkbest, target)
            _check_data(self, sk_X_new, rasl_X_new, target)


class TestOneHotEncoder(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.targets = ["pandas", "spark"]

    def test_1(self):
        """
                From https://scikit-learn.org/1.1/modules/generated/sklearn.preprocessing.OneHotEncoder.html
                >>> enc = OneHotEncoder(handle_unknown='ignore')
                >>> X = [['Male', 1], ['Female', 3], ['Female', 2]]
                >>> enc.fit(X)
                OneHotEncoder(handle_unknown='ignore')
                >>> enc.categories_
                [array(['Female', 'Male'], dtype=object), array([1, 2, 3], dtype=object)]
                >>> enc.transform([['Female', 1], ['Male', 4]]).toarray()
                array([[1., 0., 1., 0., 0.],
                    [0., 1., 0., 0., 0.]])
                >>> enc.inverse_transform([[0, 1, 1, 0, 0], [0, 0, 0, 1, 0]])
                array([['Male', 1],
                    [None, 2]], dtype=object)
                >>> enc.get_feature_names_out(['gender', 'group'])
        array(['gender_Female', 'gender_Male', 'group_1', 'group_2', 'group_3'], ...)
        """
        sk_enc = SkOneHotEncoder(handle_unknown="ignore")
        X = [["Male", 1], ["Female", 3], ["Female", 2]]
        sk_enc.fit(X)
        data = [["Female", 1], ["Male", 4]]
        sk_transformed_data = sk_enc.transform(data).toarray()
        # enc.inverse_transform([[0, 1, 1, 0, 0], [0, 0, 0, 1, 0]])

        for target in self.targets:
            X = Convert(astype=target).transform(X)
            data = Convert(astype=target).transform(data)
            rasl_enc = RaslOneHotEncoder(handle_unknown="ignore")
            rasl_enc.fit(X)
            rasl_transformed_data = rasl_enc.transform(data)
            _check_trained_one_hot_encoder(self, sk_enc, rasl_enc, target)
            _check_data(self, sk_transformed_data, rasl_transformed_data, target)
            # for i in range(len(sk_transformed_data)):
            #     for j in range(len(sk_transformed_data[i])):
            #         self.assertEqual(sk_transformed_data[i][j], rasl_transformed_data[i][j])

    def test_2(self):
        """
        From https://scikit-learn.org/1.1/modules/generated/sklearn.preprocessing.OneHotEncoder.html
        >>> drop_enc = OneHotEncoder(drop='first').fit(X)
        >>> drop_enc.categories_
        [array(['Female', 'Male'], dtype=object), array([1, 2, 3], dtype=object)]
        >>> drop_enc.transform([['Female', 1], ['Male', 2]]).toarray()
        array([[0., 0., 0.],
            [1., 1., 0.]])
        """
        pass  # XXX `drop='first'` is not supported

    def test_3(self):
        """
        From https://scikit-learn.org/1.1/modules/generated/sklearn.preprocessing.OneHotEncoder.html
        >>> drop_binary_enc = OneHotEncoder(drop='if_binary').fit(X)
        >>> drop_binary_enc.transform([['Female', 1], ['Male', 2]]).toarray()
        array([[0., 1., 0., 0.],
            [1., 0., 1., 0.]])
        """
        pass  # XXX `drop='if_binary'` is not supported

    def test_4(self):
        """
        From https://scikit-learn.org/1.1/modules/generated/sklearn.preprocessing.OneHotEncoder.html
        >>> import numpy as np
        >>> X = np.array([["a"] * 5 + ["b"] * 20 + ["c"] * 10 + ["d"] * 3], dtype=object).T
        >>> ohe = OneHotEncoder(max_categories=3, sparse=False).fit(X)
        >>> ohe.infrequent_categories_
        [array(['a', 'd'], dtype=object)]
        >>> ohe.transform([["a"], ["b"]])
        array([[0., 0., 1.],
            [1., 0., 0.]])
        """
        pass  # XXX `max_categories=3` is not supported


class TestSimpleImputer(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.targets = ["pandas", "spark"]

    def test_1(self):
        """
        From https://scikit-learn.org/1.1/modules/generated/sklearn.impute.SimpleImputer.html
        >>> import numpy as np
        >>> from sklearn.impute import SimpleImputer
        >>> imp_mean = SimpleImputer(missing_values=np.nan, strategy='mean')
        >>> imp_mean.fit([[7, 2, 3], [4, np.nan, 6], [10, 5, 9]])
        SimpleImputer()
        >>> X = [[np.nan, 2, 3], [4, np.nan, 6], [10, np.nan, 9]]
        >>> print(imp_mean.transform(X))
        [[ 7.   2.   3. ]
        [ 4.   3.5  6. ]
        [10.   3.5  9. ]]
        """
        sk_imp_mean = SkSimpleImputer(missing_values=np.nan, strategy="mean")
        training = [[7, 2, 3], [4, np.nan, 6], [10, 5, 9]]
        sk_imp_mean.fit(training)

        X = [[np.nan, 2, 3], [4, np.nan, 6], [10, np.nan, 9]]
        sk_transformed = sk_imp_mean.transform(X)
        for target in self.targets:
            training = Convert(astype=target).transform(training)
            X = Convert(astype=target).transform(X)
            rasl_imp_mean = RaslSimpleImputer(missing_values=np.nan, strategy="mean")
            rasl_imp_mean.fit(training)
            rasl_transformed = rasl_imp_mean.transform(X)
            _check_trained_simple_imputer(self, sk_imp_mean, rasl_imp_mean, target)
            _check_data(self, sk_transformed, rasl_transformed, target)


class TestStandardScaler(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.targets = ["pandas", "spark"]

    def test_1(self):
        """
        From https://scikit-learn.org/1.1/modules/generated/sklearn.preprocessing.StandardScaler.html
        >>> from sklearn.preprocessing import StandardScaler
        >>> data = [[0, 0], [0, 0], [1, 1], [1, 1]]
        >>> scaler = StandardScaler()
        >>> print(scaler.fit(data))
        StandardScaler()
        >>> print(scaler.mean_)
        [0.5 0.5]
        >>> print(scaler.transform(data))
        [[-1. -1.]
        [-1. -1.]
        [ 1.  1.]
        [ 1.  1.]]
        >>> print(scaler.transform([[2, 2]]))
        [[3. 3.]]
        """
        data = [[0, 0], [0, 0], [1, 1], [1, 1]]
        sk_scaler = SkStandardScaler()
        sk_scaler.fit(data)
        sk_transformed_data = sk_scaler.transform(data)
        data2 = [[2, 2]]
        sk_transformed_data2 = sk_scaler.transform(data2)
        for target in self.targets:
            data = Convert(astype=target).transform(data)
            data2 = Convert(astype=target).transform(data2)
            rasl_scaler = RaslStandardScaler()
            rasl_scaler.fit(data)
            rasl_transformed_data = rasl_scaler.transform(data)
            rasl_transformed_data2 = rasl_scaler.transform(data2)
            _check_trained_standard_scaler(self, sk_scaler, rasl_scaler, target)
            _check_data(self, sk_transformed_data, rasl_transformed_data, target)
            _check_data(self, sk_transformed_data2, rasl_transformed_data2, target)


class TestTargetEncoder(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.targets = ["pandas"]
        # cls.targets = ["pandas", "spark"]

    def test_1(self):
        """
        From https://contrib.scikit-learn.org/category_encoders/targetencoder.html
        >>> from category_encoders import *
        >>> import pandas as pd
        >>> from sklearn.datasets import fetch_openml
        >>> display_cols = ["Id", "MSSubClass", "MSZoning", "LotFrontage", "YearBuilt", "Heating", "CentralAir"]
        >>> bunch = fetch_openml(name="house_prices", as_frame=True)
        >>> y = bunch.target > 200000
        >>> X = pd.DataFrame(bunch.data, columns=bunch.feature_names)[display_cols]
        >>> enc = TargetEncoder(cols=['CentralAir', 'Heating'], min_samples_leaf=20, smoothing=10).fit(X, y)
        >>> numeric_dataset = enc.transform(X)
        """
        display_cols = [
            "Id",
            "MSSubClass",
            "MSZoning",
            "LotFrontage",
            "YearBuilt",
            "Heating",
            "CentralAir",
        ]
        bunch = fetch_openml(name="house_prices", as_frame=True)
        y = bunch.target > 200000
        X = pd.DataFrame(bunch.data, columns=bunch.feature_names)[display_cols]
        sk_enc = SkTargetEncoder(
            cols=["CentralAir", "Heating"], min_samples_leaf=20, smoothing=10
        ).fit(X, y)
        sk_transformed = sk_enc.transform(X)
        classes = sorted(list(y.unique()))
        for target in self.targets:
            rasl_enc = RaslTargetEncoder(
                cols=["CentralAir", "Heating"],
                min_samples_leaf=20,
                smoothing=10,
                classes=classes,
            ).fit(X, y)
            X2 = Convert(astype=target).transform(X)
            rasl_transformed = rasl_enc.transform(X2)
            _check_trained_target_encoder(self, sk_enc, rasl_enc.impl, target)
            _check_data(self, sk_transformed, rasl_transformed, target)

    def test_2(self):
        """
        >>> from category_encoders.datasets import load_compass
        >>> X, y = load_compass()
        >>> hierarchical_map = {'compass': {'N': ('N', 'NE'), 'S': ('S', 'SE'), 'W': 'W'}}
        >>> enc = TargetEncoder(verbose=1, smoothing=2, min_samples_leaf=2, hierarchy=hierarchical_map, cols=['compass']).fit(X.loc[:,['compass']], y)
        >>> hierarchy_dataset = enc.transform(X.loc[:,['compass']])
        >>> print(hierarchy_dataset['compass'].values)
        [0.62263617 0.62263617 0.90382995 0.90382995 0.90382995 0.17660024
        0.17660024 0.46051953 0.46051953 0.46051953 0.46051953 0.40332791
        0.40332791 0.40332791 0.40332791 0.40332791]
        >>> X, y = load_postcodes('binary')
        >>> cols = ['postcode']
        >>> HIER_cols = ['HIER_postcode_1','HIER_postcode_2','HIER_postcode_3','HIER_postcode_4']
        >>> enc = TargetEncoder(verbose=1, smoothing=2, min_samples_leaf=2, hierarchy=X[HIER_cols], cols=['postcode']).fit(X['postcode'], y)
        >>> hierarchy_dataset = enc.transform(X['postcode'])
        >>> print(hierarchy_dataset.loc[0:10, 'postcode'].values)
        [0.75063473 0.90208756 0.88328833 0.77041254 0.68891504 0.85012847
        0.76772574 0.88742357 0.7933824  0.63776756 0.9019973 ]
        """
        hierarchical_map = {"compass": {"N": ("N", "NE"), "S": ("S", "SE"), "W": "W"}}
        # TODO: turn this into a real test if/when hierarchy implemented
        with self.assertRaises(jsonschema.ValidationError):
            _ = RaslTargetEncoder(
                verbose=1,
                smoothing=2,
                min_samples_leaf=2,
                hierarchy=hierarchical_map,
                cols=["compass"],
            )



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Schema-enhanced versions of the operators from `autoai_libs`_ to enable hyperparameter tuning.

.. _`autoai_libs`: https://pypi.org/project/autoai-libs

Operators
=========

Preprocessing any columns:

* lale.lib.autoai_libs. `ColumnSelector`_
* lale.lib.autoai_libs. `NumpyColumnSelector`_
* lale.lib.autoai_libs. `NumpyReplaceMissingValues`_
* lale.lib.autoai_libs. `float32_transform`_
* lale.lib.autoai_libs. `NumpyPermuteArray`_

Preprocessing categorical columns:

* lale.lib.autoai_libs. `CompressStrings`_
* lale.lib.autoai_libs. `NumpyReplaceUnknownValues`_
* lale.lib.autoai_libs. `boolean2float`_
* lale.lib.autoai_libs. `CatImputer`_
* lale.lib.autoai_libs. `CatEncoder`_

Preprocessing numeric columns:

* lale.lib.autoai_libs. `FloatStr2Float`_
* lale.lib.autoai_libs. `NumImputer`_
* lale.lib.autoai_libs. `OptStandardScaler`_

Preprocessing text columns:

* lale.lib.autoai_libs. `TextTransformer`_
* lale.lib.autoai_libs. `Word2VecTransformer`_

Preprocessing date columns:

* lale.lib.autoai_libs. `DateTransformer`_

Feature transformation:

* lale.lib.autoai_libs. `TNoOp`_
* lale.lib.autoai_libs. `TA1`_
* lale.lib.autoai_libs. `TA2`_
* lale.lib.autoai_libs. `TB1`_
* lale.lib.autoai_libs. `TB2`_
* lale.lib.autoai_libs. `TAM`_
* lale.lib.autoai_libs. `TGen`_
* lale.lib.autoai_libs. `FS1`_
* lale.lib.autoai_libs. `FS2`_

.. _`ColumnSelector`: lale.lib.autoai_libs.column_selector.html
.. _`NumpyColumnSelector`: lale.lib.autoai_libs.numpy_column_selector.html
.. _`CompressStrings`: lale.lib.autoai_libs.compress_strings.html
.. _`NumpyReplaceMissingValues`: lale.lib.autoai_libs.numpy_replace_missing_values.html
.. _`NumpyReplaceUnknownValues`: lale.lib.autoai_libs.numpy_replace_unknown_values.html
.. _`boolean2float`: lale.lib.autoai_libs.boolean2float.html
.. _`CatImputer`: lale.lib.autoai_libs.cat_imputer.html
.. _`CatEncoder`: lale.lib.autoai_libs.cat_encoder.html
.. _`float32_transform`: lale.lib.autoai_libs.float32_transform.html
.. _`FloatStr2Float`: lale.lib.autoai_libs.float_str2_float.html
.. _`NumImputer`: lale.lib.autoai_libs.num_imputer.html
.. _`OptStandardScaler`: lale.lib.autoai_libs.opt_standard_scaler.html
.. _`TextTransformer`: lale.lib.autoai_libs.text_transformer.html
.. _`Word2VecTransformer`: lale.lib.autoai_libs.word2vec_transformer.html
.. _`DateTransformer`: lale.lib.autoai_libs.date_transformer.html
.. _`NumpyPermuteArray`: lale.lib.autoai_libs.numpy_permute_array.html
.. _`TNoOp`: lale.lib.autoai_libs.t_no_op.html
.. _`TA1`: lale.lib.autoai_libs.ta1.html
.. _`TA2`: lale.lib.autoai_libs.ta2.html
.. _`TB1`: lale.lib.autoai_libs.tb1.html
.. _`TB2`: lale.lib.autoai_libs.tb2.html
.. _`TAM`: lale.lib.autoai_libs.tam.html
.. _`TGen`: lale.lib.autoai_libs.tgen.html
.. _`FS1`: lale.lib.autoai_libs.fs1.html
.. _`FS2`: lale.lib.autoai_libs.fs2.html
"""

from lale import register_lale_wrapper_modules

from .boolean2float import boolean2float as boolean2float
from .cat_encoder import CatEncoder as CatEncoder
from .cat_imputer import CatImputer as CatImputer
from .column_selector import ColumnSelector as ColumnSelector
from .compress_strings import CompressStrings as CompressStrings
from .date_transformer import DateTransformer as DateTransformer
from .float32_transform import float32_transform as float32_transform
from .float_str2_float import FloatStr2Float as FloatStr2Float
from .fs1 import FS1 as FS1
from .fs2 import FS2 as FS2
from .num_imputer import NumImputer as NumImputer
from .numpy_column_selector import NumpyColumnSelector as NumpyColumnSelector
from .numpy_permute_array import NumpyPermuteArray as NumpyPermuteArray
from .numpy_replace_missing_values import (
    NumpyReplaceMissingValues as NumpyReplaceMissingValues,
)
from .numpy_replace_unknown_values import (
    NumpyReplaceUnknownValues as NumpyReplaceUnknownValues,
)
from .opt_standard_scaler import OptStandardScaler as OptStandardScaler
from .t_no_op import TNoOp as TNoOp
from .ta1 import TA1 as TA1
from .ta2 import TA2 as TA2
from .tam import TAM as TAM
from .tb1 import TB1 as TB1
from .tb2 import TB2 as TB2
from .text_transformer import TextTransformer as TextTransformer
from .tgen import TGen as TGen
from .util import wrap_pipeline_segments as wrap_pipeline_segments
from .word2vec_transformer import Word2VecTransformer as Word2VecTransformer

# Note: all imports should be done as
# from .xxx import XXX as XXX
# this ensures that pyright considers them to be publicly available
# and not private imports (this affects lale users that use pyright)


register_lale_wrapper_modules(__name__)



# Copyright 2022 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import inspect
from typing import Any, Tuple

import pandas as pd
import sklearn.base

import lale.docstrings
import lale.operators
from lale.lib._common_schemas import schema_estimator
from lale.lib.sklearn import DecisionTreeClassifier, bagging_classifier

from .monoid import Monoid, MonoidableOperator


class _BatchedBaggingClassifierMonoid(Monoid):
    def __init__(self, classifiers):
        self.classifiers = classifiers

    def combine(self, other: "_BatchedBaggingClassifierMonoid"):
        orig_classifiers = copy.copy(self.classifiers)
        orig_classifiers.extend(other.classifiers)
        return _BatchedBaggingClassifierMonoid(classifiers=orig_classifiers)


class _BatchedBaggingClassifierImpl(
    MonoidableOperator[_BatchedBaggingClassifierMonoid]
):
    def __init__(self, base_estimator=None):
        if base_estimator is None:
            base_estimator = DecisionTreeClassifier()
        self._hyperparams = {"base_estimator": base_estimator}

    def predict(self, X):
        if len(self.classifiers_list) == 1:
            return self.classifiers_list[0].predict(X)
        else:
            # Take a voting of the classifiers
            predictions_list = [
                classifier.predict(X) for classifier in self.classifiers_list
            ]
            df = pd.DataFrame(predictions_list).transpose()
            predictions = df.mode(axis=1)
            if (
                predictions.shape[1] > 1
            ):  # When there are multiple modes, pick the first one
                predictions = predictions.iloc[:, 0]
            predictions = predictions.squeeze()  # converts a dataframe to series.
            return predictions

    def from_monoid(self, monoid: _BatchedBaggingClassifierMonoid):
        self._monoid = monoid
        self.classifiers_list = monoid.classifiers

    def to_monoid(self, batch: Tuple[Any, Any]) -> _BatchedBaggingClassifierMonoid:
        X, y = batch
        trainable = self._hyperparams["base_estimator"]
        if isinstance(trainable, sklearn.base.BaseEstimator):
            trainable = sklearn.base.clone(trainable)
        if inspect.isclass(trainable):
            trainable = trainable()
        trained_classifier = trainable.fit(X, y)
        return _BatchedBaggingClassifierMonoid([trained_classifier])


_hyperparams_schema = {
    "allOf": [
        {
            "type": "object",
            "required": [
                "base_estimator",
            ],
            "relevantToOptimizer": [],
            "additionalProperties": False,
            "properties": {
                "base_estimator": schema_estimator,
            },
        }
    ]
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Implementation of a homomorphic bagging classifier.
    As proposed in https://izbicki.me/public/papers/icml2013-algebraic-classifiers.pdf""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.bagging_monoid_classifier.html",
    "type": "object",
    "tags": {
        "pre": ["~categoricals"],
        "op": ["estimator"],
        "post": [],
    },
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": bagging_classifier._input_fit_schema,
        "input_predict": bagging_classifier.schema_X_numbers,
        "output_predict": bagging_classifier.schema_1D_cats,
    },
}

BatchedBaggingClassifier = lale.operators.make_operator(
    _BatchedBaggingClassifierImpl, _combined_schemas
)
lale.docstrings.set_docstrings(BatchedBaggingClassifier)



# Copyright 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing
from typing import Any, Tuple

import numpy as np

import lale.docstrings
import lale.helpers
import lale.operators
from lale.expressions import it
from lale.expressions import sum as agg_sum
from lale.lib.dataframe import count, get_columns
from lale.lib.sklearn import standard_scaler

from .aggregate import Aggregate
from .map import Map
from .monoid import Monoid, MonoidableOperator


def scale(X, **kwargs):
    return StandardScaler(**kwargs).fit(X).transform(X)


class _StandardScalerMonoid(Monoid):
    def __init__(self, *, feature_names_in_, n_samples_seen_, _sum1, _sum2):
        self.feature_names_in_ = feature_names_in_
        self.n_samples_seen_ = n_samples_seen_
        self._sum1 = _sum1
        self._sum2 = _sum2

    def combine(self, other: "_StandardScalerMonoid"):
        assert list(self.feature_names_in_) == list(other.feature_names_in_)
        combined_feat = self.feature_names_in_
        combined_n_samples_seen = self.n_samples_seen_ + other.n_samples_seen_
        if self._sum1 is None:
            combined_sum1 = None
        else:
            assert other._sum1 is not None and len(self._sum1) == len(other._sum1)
            combined_sum1 = self._sum1 + other._sum1
        if self._sum2 is None:
            combined_sum2 = None
        else:
            assert other._sum2 is not None and len(self._sum2) == len(other._sum2)
            combined_sum2 = self._sum2 + other._sum2
        return _StandardScalerMonoid(
            feature_names_in_=combined_feat,
            n_samples_seen_=combined_n_samples_seen,
            _sum1=combined_sum1,
            _sum2=combined_sum2,
        )


class _StandardScalerImpl(MonoidableOperator[_StandardScalerMonoid]):
    def __init__(self, *, copy=True, with_mean=True, with_std=True):
        self._hyperparams = {"copy": copy, "with_mean": with_mean, "with_std": with_std}
        self.with_mean = with_mean

    def transform(self, X, copy=None):
        if self._transformer is None:
            self._transformer = self._build_transformer()
        return self._transformer.transform(X)

    def get_feature_names_out(self, input_features):
        assert input_features == self.feature_names_in_
        return self.feature_names_in_

    @property
    def n_samples_seen_(self):
        return getattr(self._monoid, "n_samples_seen_", 0)

    @property
    def feature_names_in_(self):
        return getattr(self._monoid, "feature_names_in_", None)

    def from_monoid(self, monoid: _StandardScalerMonoid):
        self._monoid = monoid
        n = monoid.n_samples_seen_
        if self._hyperparams["with_std"]:
            # Table 1 of http://www.vldb.org/pvldb/vol8/p702-tangwongsan.pdf
            self.var_ = (monoid._sum2 - monoid._sum1 * monoid._sum1 / n) / n
            self.scale_ = np.where(self.var_ == 0.0, 1.0, np.sqrt(self.var_))
        else:
            self.var_ = None
            self.scale_ = None
        if self._hyperparams["with_mean"]:
            self.mean_ = monoid._sum1 / n
        else:
            self.mean_ = None
        self.n_features_in_ = len(monoid.feature_names_in_)
        self._transformer = None

    def _build_transformer(self):
        def scale_expr(col_idx, col_name):
            expr = it[col_name]
            if self.mean_ is not None:
                expr = expr - self.mean_[col_idx]
            if self.scale_ is not None:
                expr = expr / self.scale_[col_idx]
            return expr

        assert self._monoid is not None
        result = Map(
            columns={
                col_name: scale_expr(col_idx, col_name)
                for col_idx, col_name in enumerate(self._monoid.feature_names_in_)
            }
        )
        return result

    def to_monoid(self, batch: Tuple[Any, Any]):
        X, _ = batch
        hyperparams = self._hyperparams
        feature_names_in = get_columns(X)
        n_samples_seen = count(X)
        if hyperparams["with_mean"] or hyperparams["with_std"]:
            sum1_op = Aggregate(columns={c: agg_sum(it[c]) for c in feature_names_in})
            sum1_data = lale.helpers._ensure_pandas(sum1_op.transform(X))
            sum1 = sum1_data[feature_names_in].values[0]
        else:
            sum1 = None
        if hyperparams["with_std"]:
            sum2_op = Map(
                columns={c: it[c] * it[c] for c in feature_names_in}
            ) >> Aggregate(columns={c: agg_sum(it[c]) for c in feature_names_in})
            sum2_data = lale.helpers._ensure_pandas(sum2_op.transform(X))
            sum2 = sum2_data[feature_names_in].values[0]
        else:
            sum2 = None
        return _StandardScalerMonoid(
            feature_names_in_=feature_names_in,
            n_samples_seen_=n_samples_seen,
            _sum1=sum1,
            _sum2=sum2,
        )


_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Relational algebra reimplementation of scikit-learn's `StandardScaler`_ transformer that standardizes features by removing the mean and scaling to unit variance.
Works on both pandas and Spark dataframes by using `Aggregate`_ for `fit` and `Map`_ for `transform`, which in turn use the appropriate backend.

.. _`StandardScaler`: https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.StandardScaler.html
.. _`Aggregate`: https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.aggregate.html
.. _`Map`: https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.map.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.standard_scaler.html",
    "type": "object",
    "tags": {
        "pre": ["~categoricals"],
        "op": ["transformer", "interpretable"],
        "post": [],
    },
    "properties": {
        "hyperparams": standard_scaler._hyperparams_schema,
        "input_fit": standard_scaler._input_fit_schema,
        "input_transform": standard_scaler._input_transform_schema,
        "output_transform": standard_scaler._output_transform_schema,
    },
}

StandardScaler = lale.operators.make_operator(_StandardScalerImpl, _combined_schemas)

StandardScaler = typing.cast(
    lale.operators.PlannedIndividualOp,
    StandardScaler.customize_schema(
        copy={
            "enum": [True],
            "description": "This implementation only supports `copy=True`.",
            "default": True,
        },
    ),
)

lale.docstrings.set_docstrings(StandardScaler)



# Copyright 2023 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

from lale.helpers import _is_pandas_df, _is_spark_df


# From https://github.com/scikit-learn-contrib/category_encoders/blob/master/category_encoders/utils.py
def _is_category(dtype):
    return pd.api.types.is_categorical_dtype(dtype)


# Based on https://github.com/scikit-learn-contrib/category_encoders/blob/master/category_encoders/utils.py
def get_obj_cols(df):
    """
    Returns names of 'object' columns in the DataFrame.
    """
    obj_cols = []
    if _is_pandas_df(df):
        for idx, dt in enumerate(df.dtypes):
            if dt == "object" or _is_category(dt):
                obj_cols.append(df.columns.values[idx])
    elif _is_spark_df(df):
        for idx, (col, dt) in enumerate(df.dtypes):
            if dt == "string":
                obj_cols.append(col)
    else:
        assert False
    return obj_cols



# Copyright 2023 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains common schema fragments used in the autoai_libs schemas
"""

from typing import Any, Dict

JSON_TYPE = Dict[str, Any]

# activate_flag
_hparam_activate_flag_unmodified: JSON_TYPE = {
    "description": "If False, transform(X) outputs the input numpy array X unmodified.",
    "type": "boolean",
    "default": True,
}

_hparam_activate_flag_features: JSON_TYPE = {
    "description": "If False, the features are not generated.",
    "type": "boolean",
    "default": True,
}

_hparam_activate_flag_active: JSON_TYPE = {
    "description": "Determines whether transformer is active or not.",
    "type": "boolean",
    "default": True,
}

_hparam_col_dtypes: JSON_TYPE = {
    "description": "List of the datatypes of the feature columns.",
    "anyOf": [
        {"type": "array", "items": {"laleType": "Any"}},
        {"enum": [None]},
    ],
    "default": None,
}

_hparam_dtypes_list: JSON_TYPE = {
    "anyOf": [
        {
            "description": "Strings that denote the type of each column of the input numpy array X.",
            "type": "array",
            "items": {
                "enum": [
                    "char_str",
                    "int_str",
                    "float_str",
                    "float_num",
                    "float_int_num",
                    "int_num",
                    "boolean",
                    "Unknown",
                    "missing",
                ]
            },
        },
        {
            "description": "If None, the column types are discovered.",
            "enum": [None],
        },
    ],
    "default": None,
}

_hparam_sklearn_version_family: JSON_TYPE = {
    "description": "The sklearn version for backward compatibility with versions 019 and 020dev. Currently unused.",
    "enum": ["20", "21", "22", "23", "24", None, "1"],
    "default": None,
}


def _hparam_column_headers_list(description: str) -> JSON_TYPE:
    return {
        "description": description,
        "anyOf": [
            {"type": "array", "items": {"type": "string"}},
            {"type": "array", "items": {"type": "integer"}},
            {"enum": [None]},
        ],
        "default": None,
    }


_hparam_fs_cols_ids_must_keep: JSON_TYPE = {
    "description": "Serial numbers of the columns that must be kept irrespective of their feature importance.",
    "laleType": "Any",  # Found a value `range(0, 20)`
    "default": [],
}

_hparams_fs_additional_col_count_to_keep: JSON_TYPE = {
    "description": "How many columns need to be retained.",
    "type": "integer",
    "minimum": 0,
}

_hparams_fs_ptype: JSON_TYPE = {
    "description": "Problem type.",
    "enum": ["classification", "regression"],
    "default": "classification",
}


def _hparams_column_index_list(description: str) -> JSON_TYPE:
    return {
        "description": description,
        "anyOf": [
            {"type": "array", "items": {"type": "integer", "minimum": 0}},
            {"enum": [None]},
        ],
        "default": None,
    }


_hparams_transformer_name: JSON_TYPE = {
    "description": "A string name that uniquely identifies this transformer from others.",
    "anyOf": [{"type": "string"}, {"enum": [None]}],
    "default": None,
}


def _hparams_fun_pointer(description: str) -> JSON_TYPE:
    return {
        "description": description,
        "laleType": "Any",
        "default": None,
    }


_hparams_datatype_spec: JSON_TYPE = {"type": "array", "items": {"type": "string"}}


def _hparams_datatypes(description: str) -> JSON_TYPE:
    return {
        "description": description,
        "anyOf": [
            _hparams_datatype_spec,
            {"enum": [None]},
        ],
        "default": None,
    }


def _hparams_feat_constraints(description: str) -> JSON_TYPE:
    return {
        "description": description,
        "laleType": "Any",
        "default": None,
    }


_hparams_tgraph: JSON_TYPE = {
    "description": "Should be the invoking TGraph() object.",
    "anyOf": [
        {"laleType": "Any"},
        {
            "enum": [None],
            "description": "Passing None will result in some failure to detect some inefficiencies due to lack of caching.",
        },
    ],
    "default": None,
}

_hparams_apply_all: JSON_TYPE = {
    "description": "Only use applyAll = True. It means that the transformer will enumerate all features (or feature sets) that match the specified criteria and apply the provided function to each.",
    "type": "boolean",
    "default": True,
}

_hparams_col_names: JSON_TYPE = {
    "description": "Names of the feature columns in a list.",
    "anyOf": [
        {"type": "array", "items": {"type": "string"}},
        {"enum": [None]},
    ],
    "default": None,
}

_hparams_col_as_json_objects: JSON_TYPE = {
    "description": "Names of the feature columns in a json dict.",
    "anyOf": [{"type": "object"}, {"enum": [None]}],
    "default": None,
}

_hparams_tans_class: JSON_TYPE = {
    "description": "A class that implements fit() and transform() in accordance with the transformation function definition.",
    "laleType": "Any",
    "default": None,
}



# Copyright 2022 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

import numpy as np
from packaging import version

try:
    import category_encoders

    catenc_version = version.parse(getattr(category_encoders, "__version__"))

except ImportError:
    catenc_version = None

import lale.docstrings
import lale.operators

_hyperparams_schema = {
    "allOf": [
        {
            "type": "object",
            "additionalProperties": False,
            "required": [
                "verbose",
                "cols",
                "drop_invariant",
                "return_df",
                "handle_missing",
                "handle_unknown",
                "min_samples_leaf",
                "smoothing",
            ],
            "relevantToOptimizer": [],
            "properties": {
                "verbose": {
                    "type": "integer",
                    "description": "Verbosity of the output, 0 for none.",
                    "default": 0,
                },
                "cols": {
                    "description": "Columns to encode.",
                    "anyOf": [
                        {
                            "enum": [None],
                            "description": "All string columns will be encoded.",
                        },
                        {
                            "type": "array",
                            "items": {"type": "string"},
                        },
                    ],
                    "default": None,
                },
                "drop_invariant": {
                    "type": "boolean",
                    "default": False,
                    "description": "Whether to drop columns with 0 variance.",
                },
                "return_df": {
                    "type": "boolean",
                    "default": True,
                    "description": "Whether to return a pandas DataFrame from transform (otherwise it will be a numpy array).",
                },
                "handle_missing": {
                    "enum": ["error", "return_nan", "value"],
                    "default": "value",
                    "description": "Given 'value', return the target mean.",
                },
                "handle_unknown": {
                    "enum": ["error", "return_nan", "value"],
                    "default": "value",
                    "description": "Given 'value', return the target mean.",
                },
                "min_samples_leaf": {
                    "type": "integer",
                    "default": 1,
                    "minimum": 1,
                    "maximumForOptimizer": 10,
                    "description": "For regularization the weighted average between category mean and global mean is taken. The weight is an S-shaped curve between 0 and 1 with the number of samples for a category on the x-axis. The curve reaches 0.5 at min_samples_leaf. (parameter k in the original paper)",
                },
                "smoothing": {
                    "type": "number",
                    "default": 1.0,
                    "minimum": 0.0,
                    "exclusiveMinimum": True,
                    "maximumForOptimizer": 10.0,
                    "description": "Smoothing effect to balance categorical average vs prior. Higher value means stronger regularization. The value must be strictly bigger than 0. Higher values mean a flatter S-curve (see min_samples_leaf).",
                },
            },
        }
    ],
}

_input_fit_schema = {
    "type": "object",
    "required": ["X", "y"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {
                "type": "array",
                "items": {"anyOf": [{"type": "number"}, {"type": "string"}]},
            },
        },
        "y": {
            "description": "Target class labels; the array is over samples.",
            "anyOf": [
                {"type": "array", "items": {"type": "number"}},
                {"type": "array", "items": {"type": "string"}},
            ],
        },
    },
}

_input_transform_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {
                "type": "array",
                "items": {"anyOf": [{"type": "number"}, {"type": "string"}]},
            },
        }
    },
}

_output_transform_schema = {
    "type": "array",
    "items": {"type": "array", "items": {"type": "number"}},
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Target encoder`_ transformer from scikit-learn contrib that encodes categorical features as numbers.

.. _`Target encoder`: https://contrib.scikit-learn.org/category_encoders/targetencoder.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.category_encoders.target_encoder.html",
    "import_from": "category_encoders.target_encoder",
    "type": "object",
    "tags": {"pre": ["categoricals"], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
    },
}


class _TargetEncoderImpl:
    def __init__(self, **hyperparams):
        if catenc_version is None:
            raise ValueError("The package 'category_encoders' is not installed.")
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=FutureWarning)
            self._wrapped_model = category_encoders.TargetEncoder(**hyperparams)

    def fit(self, X, y):
        if catenc_version is None:
            raise ValueError("The package 'category_encoders' is not installed.")
        if np.issubdtype(y.dtype, np.number):
            numeric_y = y
        else:
            from sklearn.preprocessing import LabelEncoder

            trainable_le = LabelEncoder()
            trained_le = trainable_le.fit(y)
            numeric_y = trained_le.transform(y)
        self._wrapped_model.fit(X, numeric_y)
        return self

    def transform(self, X):
        if catenc_version is None:
            raise ValueError("The package 'category_encoders' is not installed.")
        result = self._wrapped_model.transform(X)
        return result


TargetEncoder = lale.operators.make_operator(_TargetEncoderImpl, _combined_schemas)

if catenc_version is not None and catenc_version >= version.Version("2.5.1"):
    TargetEncoder = TargetEncoder.customize_schema(
        hierarchy={
            "laleType": "Any",
            "default": None,
            "description": "A dictionary or a dataframe to define the hierarchy for mapping.",
        },
    )

lale.docstrings.set_docstrings(TargetEncoder)



# Copyright 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numbers
import typing
from typing import Any, Tuple

import numpy as np
import pandas as pd

import lale.docstrings
import lale.operators
from lale.expressions import sum  # pylint:disable=redefined-builtin
from lale.expressions import count, it, median, mode, replace
from lale.helpers import _is_df, _is_pandas_df, _is_spark_df
from lale.lib.dataframe import get_columns
from lale.lib.sklearn import simple_imputer
from lale.schemas import Enum

from .aggregate import Aggregate
from .map import Map
from .monoid import Monoid, MonoidableOperator


def _is_numeric_df(X):
    if _is_pandas_df(X):
        return X.shape[1] == X.select_dtypes(include=np.number).shape[1]
    elif _is_spark_df(X):
        from pyspark.sql.types import NumericType

        numeric_cols = [
            f.name for f in X.schema.fields if isinstance(f.dataType, NumericType)
        ]
        for index_name in X.index_names:
            if index_name in numeric_cols:
                numeric_cols.remove(index_name)
        return len(get_columns(X)) == len(numeric_cols)
    else:
        return False


def _is_string_df(X):
    if _is_pandas_df(X):
        return X.shape[1] == X.select_dtypes(include="object").shape[1]
    elif _is_spark_df(X):
        from pyspark.sql.types import StringType

        numeric_cols = [
            f.name for f in X.schema.fields if isinstance(f.dataType, StringType)
        ]
        return len(get_columns(X)) == len(numeric_cols)
    else:
        return False


class _SimpleImputerMonoid(Monoid):
    def __init__(self, *, feature_names_in_, lifted_statistics, strategy):
        self.feature_names_in_ = feature_names_in_
        self.lifted_statistics = lifted_statistics
        self.strategy = strategy

    def combine(self, other: "_SimpleImputerMonoid"):
        assert list(self.feature_names_in_) == list(other.feature_names_in_)
        if self.strategy == "constant":
            assert self.lifted_statistics.equals(other.lifted_statistics)
            combined_statistic = self.lifted_statistics
        elif self.strategy == "mean":
            combined_statistic = {}
            combined_statistic["sum"] = (
                self.lifted_statistics["sum"] + other.lifted_statistics["sum"]
            )
            combined_statistic["count"] = (
                self.lifted_statistics["count"] + other.lifted_statistics["count"]
            )
        else:
            raise ValueError(
                "_combine is only supported for imputation strategy `mean` and `constant`."
            )
        return _SimpleImputerMonoid(
            feature_names_in_=self.feature_names_in_,
            lifted_statistics=combined_statistic,
            strategy=self.strategy,
        )


class _SimpleImputerImpl(MonoidableOperator[_SimpleImputerMonoid]):
    def __init__(
        self,
        missing_values=np.nan,
        strategy="mean",
        fill_value=None,
        verbose=0,
        copy=True,
        add_indicator=False,
    ):
        self._hyperparams = {}
        self._hyperparams["missing_values"] = missing_values
        self._hyperparams["strategy"] = strategy
        self._hyperparams["fill_value"] = fill_value
        self._hyperparams["verbose"] = verbose
        if not copy:
            raise ValueError("This implementation only supports `copy=True`.")
        self._hyperparams["copy"] = copy
        if add_indicator:
            raise ValueError("This implementation only supports `add_indicator=False`.")
        self._hyperparams["add_indicator"] = add_indicator
        # the `indicator_`` property is always None as we do not support `add_indicator=True`
        self.indicator_ = None

    def to_monoid(self, batch: Tuple[Any, Any]):
        hyperparams = self._hyperparams
        X, _ = batch
        feature_names_in_ = get_columns(X)
        agg_data = None
        # learn the values to be imputed
        strategy = hyperparams["strategy"]
        if strategy == "constant":
            fill_value = _SimpleImputerImpl._get_fill_value(X, hyperparams)
            agg_data = [[fill_value for col in get_columns(X)]]
            lifted_statistics = pd.DataFrame(agg_data, columns=get_columns(X))
        elif strategy == "mean":
            agg_op_sum = Aggregate(
                columns={c: sum(it[c]) for c in get_columns(X)},
                exclude_value=hyperparams["missing_values"],
            )
            agg_op_count = Aggregate(
                columns={c: count(it[c]) for c in get_columns(X)},
                exclude_value=hyperparams["missing_values"],
            )
            lifted_statistics = {}
            agg_sum = agg_op_sum.transform(X)
            if agg_sum is not None and _is_spark_df(agg_sum):
                agg_sum = agg_sum.toPandas()
            agg_count = agg_op_count.transform(X)
            if agg_count is not None and _is_spark_df(agg_count):
                agg_count = agg_count.toPandas()
            lifted_statistics["sum"] = agg_sum
            lifted_statistics["count"] = agg_count
        else:
            raise ValueError(
                "SimpleImputer can create a Monoind only for imputation strategy `mean` and `constant`."
            )
        return _SimpleImputerMonoid(
            feature_names_in_=feature_names_in_,
            lifted_statistics=lifted_statistics,
            strategy=strategy,
        )

    def from_monoid(self, monoid: _SimpleImputerMonoid):
        self._monoid = monoid
        self.feature_names_in_ = monoid.feature_names_in_
        self.n_features_in_ = len(self.feature_names_in_)
        _lifted_statistics = monoid.lifted_statistics
        strategy = self._hyperparams["strategy"]
        if strategy == "constant":
            self.statistics_ = _lifted_statistics.to_numpy()[0]
        elif strategy == "mean":
            self.statistics_ = (
                _lifted_statistics["sum"] / _lifted_statistics["count"]
            ).to_numpy()[0]
        self._transformer = None

    def fit(self, X, y=None):
        self._validate_input(X)

        agg_op = None
        agg_data = None
        # learn the values to be imputed
        if self._hyperparams["strategy"] in ["mean", "constant"]:
            lifted = self.to_monoid((X, y))
            self.from_monoid(lifted)
            return self
        elif self._hyperparams["strategy"] == "median":
            agg_op = Aggregate(
                columns={c: median(it[c]) for c in get_columns(X)},
                exclude_value=self._hyperparams["missing_values"],
            )
        elif self._hyperparams["strategy"] == "most_frequent":
            agg_op = Aggregate(
                columns={c: mode(it[c]) for c in get_columns(X)},
                exclude_value=self._hyperparams["missing_values"],
            )
        if agg_data is None and agg_op is not None:
            agg_data = agg_op.transform(X)
        self.feature_names_in_ = get_columns(X)
        self.n_features_in_ = len(self.feature_names_in_)
        if agg_data is not None and _is_spark_df(agg_data):
            agg_data = agg_data.toPandas()
        if agg_data is not None and _is_pandas_df(agg_data):
            self.statistics_ = agg_data.to_numpy()[
                0
            ]  # Converting from a 2-d array to 1-d
        self._transformer = None
        return self

    def _build_transformer(self):
        # prepare the transformer
        transformer = Map(
            columns={
                col_name: replace(
                    it[col_name],
                    {self._hyperparams["missing_values"]: self.statistics_[col_idx]},
                )
                for col_idx, col_name in enumerate(self.feature_names_in_)
            }
        )
        return transformer

    def transform(self, X):
        if self._transformer is None:
            self._transformer = self._build_transformer()
        return self._transformer.transform(X)

    @staticmethod
    def _get_fill_value(X, hyperparams):
        # assign appropriate value to fill_value depending on the datatype.
        # default fill_value is 0 for numerical input and "missing_value"
        # otherwise
        if hyperparams["fill_value"] is None:
            if _is_numeric_df(X):
                fill_value = 0
            else:
                fill_value = "missing_value"
        else:
            fill_value = hyperparams["fill_value"]
        # validate that fill_value is numerical for numerical data
        if (
            hyperparams["strategy"] == "constant"
            and _is_numeric_df(X)
            and not isinstance(fill_value, numbers.Real)
        ):
            raise ValueError(
                f"'fill_value'={fill_value} is invalid. Expected a "
                f"numerical value when imputing numerical "
                f"data"
            )
        return fill_value

    def _validate_input(self, X):
        # validate that the dataset is either a pandas dataframe or spark.
        # For example, sparse matrix is not allowed.
        if not _is_df(X):
            raise ValueError(
                f"""Unsupported type(X) {type(X)} for SimpleImputer.
            Only pandas.DataFrame or pyspark.sql.DataFrame are allowed."""
            )
        # validate input to check the correct dtype and strategy
        # `mean` and `median` are not applicable to string inputs
        if not _is_numeric_df(X) and self._hyperparams["strategy"] in [
            "mean",
            "median",
        ]:
            raise ValueError(
                f"Cannot use {self._hyperparams['strategy']} strategy with non-numeric data."
            )

        # Check that missing_values are the right type
        if _is_numeric_df(X) and not isinstance(
            self._hyperparams["missing_values"], numbers.Real
        ):
            raise ValueError(
                f"'X' and 'missing_values' types are expected to be"
                f" both numerical. Got X.dtypes={X.dtypes} and "
                f" type(missing_values)={type(self._hyperparams['missing_values'])}."
            )


_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Relational algebra reimplementation of scikit-learn's `SimpleImputer`_.
Works on both pandas and Spark dataframes by using `Aggregate`_ for `fit` and `Map`_ for `transform`, which in turn use the appropriate backend.

.. _`SimpleImputer`: https://scikit-learn.org/stable/modules/generated/sklearn.imputer.SimpleImputer.html
.. _`Aggregate`: https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.aggregate.html
.. _`Map`: https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.map.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.simple_imputer.html",
    "type": "object",
    "tags": {
        "pre": [],
        "op": ["transformer", "interpretable"],
        "post": [],
    },
    "properties": {
        "hyperparams": simple_imputer._hyperparams_schema,
        "input_fit": simple_imputer._input_fit_schema,
        "input_transform": simple_imputer._input_transform_schema,
        "output_transform": simple_imputer._output_transform_schema,
    },
}

SimpleImputer = lale.operators.make_operator(_SimpleImputerImpl, _combined_schemas)

SimpleImputer = typing.cast(
    lale.operators.PlannedIndividualOp,
    SimpleImputer.customize_schema(
        copy=Enum(
            values=[True],
            desc="`copy=True` is the only value currently supported by this implementation",
            default=True,
        ),
        add_indicator=Enum(
            values=[False],
            desc="`add_indicator=False` is the only value currently supported by this implementation",
            default=False,
        ),
    ),
)

lale.docstrings.set_docstrings(SimpleImputer)



# Copyright 2019-2023 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

_logger = logging.getLogger()
_old_log_level = _logger.getEffectiveLevel()
_logger.setLevel(level=logging.ERROR)

# the following triggers spurious AIF360 warning "No module named 'fairlearn'":
import aif360.algorithms.inprocessing  # isort:skip # noqa:E402,F401  # pylint:disable=wrong-import-position,wrong-import-order

# the following triggers spurious AIF360 warning "No module named 'tempeh'":
import aif360.datasets  # isort:skip # noqa:E402,F401  # pylint:disable=wrong-import-position,wrong-import-order

_logger.setLevel(_old_log_level)

dummy = "dummy"



# Copyright 2022 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import lale.docstrings
import lale.operators
from lale.datasets.data_schemas import forward_metadata, get_index_names
from lale.helpers import _is_pandas_df, _is_pandas_series, _is_spark_df


class _SortIndexImpl:
    def __init__(self, ascending=True):
        self.ascending = ascending

    def transform_schema(self, s_X):
        return s_X

    def transform(self, X):
        if _is_pandas_df(X):
            ordered_df = X.sort_index(ascending=self.ascending)
        elif _is_spark_df(X):
            index_cols = get_index_names(X)  # type:ignore
            ordered_df = X.orderBy(index_cols, ascending=self.ascending)
        else:
            raise ValueError(
                "Only Pandas or Spark dataframe are supported as inputs. Please check that pyspark is installed if you see this error for a Spark dataframe."
            )
        ordered_df = forward_metadata(X, ordered_df)
        return ordered_df

    def transform_X_y(self, X, y=None):
        result_y = None
        if y is not None:
            assert _is_pandas_df(y) or _is_pandas_series(
                y
            ), "transform_X_y is supported only when y is a Pandas Series or DataFrame."
            result_y = y.sort_index(
                ascending=self.ascending
            )  # assumes that y is always Pandas
        result_X = self.transform(X)
        return result_X, result_y


_hyperparams_schema = {
    "allOf": [
        {
            "description": "This first sub-object lists all constructor arguments with their "
            "types, one at a time, omitting cross-argument constraints, if any.",
            "type": "object",
            "additionalProperties": False,
            "required": ["ascending"],
            "relevantToOptimizer": [],
            "properties": {
                "ascending": {
                    "description": "Sort by index of the dataframe.",
                    "type": "boolean",
                    "default": True,
                }
            },
        }
    ]
}


_input_transform_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "The outer array is over rows.",
            "anyOf": [
                {"laleType": "Any"},
                {
                    "type": "array",
                    "items": {
                        "description": "The inner array is over columns.",
                        "type": "array",
                        "items": {"laleType": "Any"},
                    },
                },
            ],
        }
    },
}

_output_transform_schema = {
    "description": "The outer array is over rows.",
    "anyOf": [
        {
            "type": "array",
            "items": {
                "description": "The inner array is over columns.",
                "type": "array",
                "items": {"laleType": "Any"},
            },
        },
        {"laleType": "Any"},
    ],
}

_input_transform_X_y_schema = {
    "type": "object",
    "required": ["X", "y"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Input features as numpy, pandas, or PySpark.",
            "type": "array",
            "items": {"type": "array", "items": {"laleType": "Any"}},
        },
        "y": {
            "anyOf": [
                {"enum": [None]},
                {
                    "description": "Input labels as numpy, pandas, or PySpark.",
                    "type": "array",
                    "items": {"laleType": "Any"},
                },
            ],
        },
    },
}

_output_transform_X_y_schema = {
    "type": "array",
    "laleType": "tuple",
    "items": [
        {
            "description": "X",
            "type": "array",
            "items": {"type": "array", "items": {"laleType": "Any"}},
        },
        {
            "anyOf": [
                {"enum": [None]},
                {
                    "description": "Input labels as numpy, pandas, or PySpark.",
                    "type": "array",
                    "items": {"laleType": "Any"},
                },
            ],
        },
    ],
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "SortIndex operator.",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.sort_index.html",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
        "input_transform_X_y": _input_transform_X_y_schema,
        "output_transform_X_y": _output_transform_X_y_schema,
    },
}


SortIndex = lale.operators.make_operator(_SortIndexImpl, _combined_schemas)

lale.docstrings.set_docstrings(SortIndex)



# Copyright 2019-2023 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import aif360.algorithms.postprocessing

import lale.docstrings
import lale.operators

from .util import (
    _BasePostEstimatorImpl,
    _categorical_fairness_properties,
    _categorical_input_predict_schema,
    _categorical_output_predict_schema,
    _categorical_supervised_input_fit_schema,
)


class _RejectOptionClassificationImpl(_BasePostEstimatorImpl):
    def __init__(
        self,
        *,
        favorable_labels,
        protected_attributes,
        unfavorable_labels=None,
        estimator,
        redact=True,
        repair_level=None,
        **hyperparams,
    ):
        prot_attr_names = [pa["feature"] for pa in protected_attributes]
        unprivileged_groups = [{name: 0 for name in prot_attr_names}]
        privileged_groups = [{name: 1 for name in prot_attr_names}]
        if repair_level is not None:
            hyperparams["metric_lb"] = -(1 - repair_level)
            hyperparams["metric_ub"] = 1 - repair_level
        mitigator = aif360.algorithms.postprocessing.RejectOptionClassification(
            unprivileged_groups=unprivileged_groups,
            privileged_groups=privileged_groups,
            **hyperparams,
        )
        super().__init__(
            favorable_labels=favorable_labels,
            protected_attributes=protected_attributes,
            unfavorable_labels=unfavorable_labels,
            estimator=estimator,
            redact=redact,
            mitigator=mitigator,
        )

    def predict_proba(self, X):
        raise NotImplementedError()


_input_fit_schema = _categorical_supervised_input_fit_schema
_input_predict_schema = _categorical_input_predict_schema
_output_predict_schema = _categorical_output_predict_schema

_hyperparams_schema = {
    "description": "Hyperparameter schema.",
    "allOf": [
        {
            "description": "This first sub-object lists all constructor arguments with their "
            "types, one at a time, omitting cross-argument constraints.",
            "type": "object",
            "additionalProperties": False,
            "required": [
                *_categorical_fairness_properties.keys(),
                "estimator",
                "redact",
                "low_class_thresh",
                "high_class_thresh",
                "num_class_thresh",
                "num_ROC_margin",
                "metric_name",
                "metric_ub",
                "metric_lb",
            ],
            "relevantToOptimizer": ["metric_name"],
            "properties": {
                **_categorical_fairness_properties,
                "estimator": {
                    "description": "Nested supervised learning operator for which to mitigate fairness.",
                    "laleType": "operator",
                },
                "redact": {
                    "description": "Whether to redact protected attributes before data preparation (recommended) or not.",
                    "type": "boolean",
                    "default": True,
                },
                "low_class_thresh": {
                    "description": "Smallest classification threshold to use in the optimization.",
                    "type": "number",
                    "minimum": 0.0,
                    "maximum": 1.0,
                    "default": 0.01,
                },
                "high_class_thresh": {
                    "description": "Highest classification threshold to use in the optimization.",
                    "type": "number",
                    "minimum": 0.0,
                    "maximum": 1.0,
                    "default": 0.99,
                },
                "num_class_thresh": {
                    "description": "Number of classification thresholds between low_class_thresh and high_class_thresh for the optimization search.",
                    "type": "integer",
                    "minimum": 1,
                    "default": 100,
                },
                "num_ROC_margin": {
                    "description": "Number of relevant ROC margins to be used in the optimization search.",
                    "type": "integer",
                    "minimum": 1,
                    "default": 50,
                },
                "metric_name": {
                    "description": "Name of the metric to use for the optimization.",
                    "enum": [
                        "Statistical parity difference",
                        "Average odds difference",
                        "Equal opportunity difference",
                    ],
                    "default": "Statistical parity difference",
                },
                "metric_ub": {
                    "description": "Upper bound of constraint on the metric value.",
                    "type": "number",
                    "minimum": 0,
                    "default": 0.05,
                    "maximum": 1,
                },
                "metric_lb": {
                    "description": "Lower bound of constraint on the metric value.",
                    "type": "number",
                    "minimum": -1,
                    "default": -0.05,
                    "maximum": 0,
                },
                "repair_level": {
                    "description": "Repair amount from 0 = none to 1 = full.",
                    "anyOf": [
                        {
                            "description": "Keep metric_lb and metric_ub unchanged.",
                            "enum": [None],
                        },
                        {
                            "description": "Set metric_ub = 1 - repair_level and metric_lb = - metric_ub.",
                            "type": "number",
                            "minimum": 0,
                            "maximum": 1,
                        },
                    ],
                    "default": None,
                },
            },
        }
    ],
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Reject option classification`_ post-estimator fairness mitigator. Gives favorable outcomes to unpriviliged groups and unfavorable outcomes to priviliged groups in a confidence band around the decision boundary with the highest uncertainty (`Kamiran et al. 2012`_).

.. _`Reject option classification`: https://aif360.readthedocs.io/en/latest/modules/generated/aif360.algorithms.postprocessing.RejectOptionClassification.html
.. _`Kamiran et al. 2012`: https://doi.org/10.1109/ICDM.2012.45
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.aif360.reject_option_classification.html#lale.lib.aif360.reject_option_classification.RejectOptionClassification",
    "import_from": "aif360.algorithms.postprocessing",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "classifier", "interpretable"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
    },
}


RejectOptionClassification = lale.operators.make_operator(
    _RejectOptionClassificationImpl, _combined_schemas
)

lale.docstrings.set_docstrings(RejectOptionClassification)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, List, Optional, Tuple, Type, Union

import numpy as np
from numpy import issubdtype, ndarray
from pandas import DataFrame, Series
from pandas.core.groupby import DataFrameGroupBy, SeriesGroupBy
from scipy.sparse import csr_matrix

import lale.type_checking
from lale.helpers import _is_spark_df
from lale.type_checking import JSON_TYPE

try:
    import torch
    from torch import Tensor

    torch_installed = True
except ImportError:
    torch_installed = False

try:
    import py4j.protocol
    from pyspark.sql import DataFrame as SparkDataFrame
    from pyspark.sql import GroupedData as SparkGroupedData

    spark_installed = True
except ImportError:
    spark_installed = False


# See instructions for subclassing numpy ndarray:
# https://docs.scipy.org/doc/numpy/user/basics.subclassing.html
class NDArrayWithSchema(ndarray):
    def __new__(
        cls,
        shape,
        dtype=float,
        buffer=None,
        offset=0,
        strides=None,
        order=None,
        json_schema=None,
        table_name=None,
    ):
        result = super(  # pylint:disable=too-many-function-args
            NDArrayWithSchema, cls
        ).__new__(
            cls, shape, dtype, buffer, offset, strides, order  # type: ignore
        )
        result.json_schema = json_schema
        result.table_name = table_name
        return result

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.json_schema = getattr(obj, "json_schema", None)
        self.table_name = getattr(obj, "table_name", None)


# See instructions for subclassing pandas DataFrame:
# https://pandas.pydata.org/pandas-docs/stable/development/extending.html#extending-subclassing-pandas
class DataFrameWithSchema(DataFrame):
    _internal_names = DataFrame._internal_names + ["json_schema", "table_name"]
    _internal_names_set = set(_internal_names)

    @property
    def _constructor(self):
        return DataFrameWithSchema


class SeriesWithSchema(Series):
    _internal_names = DataFrame._internal_names + [
        "json_schema",
        "table_name",
        "folds_for_monoid",
    ]
    _internal_names_set = set(_internal_names)

    @property
    def _constructor(self):
        return SeriesWithSchema


if spark_installed:

    def _gen_index_name(df, cpt=None):
        name = f"index{cpt if cpt is not None else ''}"
        if name in df.columns:
            return _gen_index_name(df, cpt=cpt + 1 if cpt is not None else 0)
        else:
            return name

    class SparkDataFrameWithIndex(SparkDataFrame):  # type: ignore
        def __init__(self, df, index_names=None):
            if index_names is not None and len(index_names) == 1:
                index_name = index_names[0]
            elif index_names is None or len(index_names) == 0:
                index_name = _gen_index_name(df)
                index_names = [index_name]
            else:
                index_name = None
            if index_name is not None and index_name not in df.columns:
                df_with_index = (
                    df.rdd.zipWithIndex()
                    .map(lambda row: row[0] + (row[1],))
                    .toDF(df.columns + [index_name])
                )
            else:
                df_with_index = df
            table_name = get_table_name(df)
            if table_name is not None:
                df_with_index = df_with_index.alias(table_name)
            super().__init__(df_with_index._jdf, df_with_index.sql_ctx)
            self.index_name = index_name
            self.index_names = index_names
            for f in df.schema.fieldNames():
                self.schema[f].metadata = df.schema[f].metadata

        def drop_indexes(self):
            result = self.drop(*self.index_names)
            result = add_table_name(result, get_table_name(self))
            return result

        @property
        def columns_without_indexes(self):
            cols = list(self.columns)
            for name in self.index_names:
                cols.remove(name)
            return cols

        def toPandas(self, *args, **kwargs):
            df = super().toPandas(*args, **kwargs)
            return df.set_index(self.index_names)

else:

    class SparkDataFrameWithIndex:  # type: ignore
        def __init__(self, df, index_names=None) -> None:
            raise ValueError("pyspark is not installed")  # type: ignore

        @property
        def index_name(self) -> Union[str, None]:
            raise ValueError("pyspark is not installed")  # type: ignore

        @property
        def index_names(self) -> List[str]:
            raise ValueError("pyspark is not installed")  # type: ignore

        def toPandas(self, *args, **kwargs) -> DataFrame:
            raise ValueError("pyspark is not installed")  # type: ignore

        @property
        def schema(self) -> Any:
            raise ValueError("pyspark is not installed")  # type: ignore


def add_schema(obj, schema=None, raise_on_failure=False, recalc=False) -> Any:
    from lale.settings import disable_data_schema_validation

    if disable_data_schema_validation:
        return obj
    if obj is None:
        return None
    if isinstance(obj, NDArrayWithSchema):
        result = obj
    elif isinstance(obj, ndarray):
        result = obj.view(NDArrayWithSchema)
    elif isinstance(obj, SeriesWithSchema):
        result = obj
    elif isinstance(obj, Series):
        result = SeriesWithSchema(obj)
    elif isinstance(obj, DataFrameWithSchema):
        result = obj
    elif isinstance(obj, DataFrame):
        result = DataFrameWithSchema(obj)
    elif is_list_tensor(obj):
        obj = np.array(obj)
        result = obj.view(NDArrayWithSchema)
    elif raise_on_failure:
        raise ValueError(f"unexpected type(obj) {type(obj)}")
    else:
        return obj
    if recalc:
        setattr(result, "json_schema", None)
    if getattr(result, "json_schema", None) is None:
        if schema is None:
            setattr(result, "json_schema", to_schema(obj))
        else:
            lale.type_checking.validate_is_schema(schema)
            setattr(result, "json_schema", schema)
    return result


def add_schema_adjusting_n_rows(obj, schema):
    assert isinstance(obj, (ndarray, DataFrame, Series)), type(obj)
    assert schema.get("type", None) == "array", schema
    n_rows = obj.shape[0]
    mod_schema = {**schema, "minItems": n_rows, "maxItems": n_rows}
    result = add_schema(obj, mod_schema)
    return result


def add_table_name(obj, name) -> Any:
    if obj is None:
        return None
    if name is None:
        return obj
    if spark_installed and isinstance(obj, SparkDataFrame):
        # alias method documentation: https://spark.apache.org/docs/latest/api/python/reference/api/pyspark.sql.DataFrame.alias.html
        # Python class DataFrame with method alias(self, alias): https://github.com/apache/spark/blob/master/python/pyspark/sql/dataframe.py
        # Scala type DataFrame: https://github.com/apache/spark/blob/master/sql/core/src/main/scala/org/apache/spark/sql/package.scala
        # Scala class DataSet with method as(alias: String): https://github.com/apache/spark/blob/master/sql/core/src/main/scala/org/apache/spark/sql/Dataset.scala
        o = obj.alias(name)
        for f in obj.schema.fieldNames():
            o.schema[f].metadata = obj.schema[f].metadata
        if isinstance(obj, SparkDataFrameWithIndex):
            o = SparkDataFrameWithIndex(o, obj.index_names)
        return o
    if isinstance(obj, NDArrayWithSchema):
        result = obj.view(NDArrayWithSchema)
        if hasattr(obj, "json_schema"):
            result.json_schema = obj.json_schema
    elif isinstance(obj, ndarray):
        result = obj.view(NDArrayWithSchema)
    elif isinstance(obj, SeriesWithSchema):
        result = obj.copy(deep=False)
        if hasattr(obj, "json_schema"):
            result.json_schema = obj.json_schema
    elif isinstance(obj, Series):
        result = SeriesWithSchema(obj)
    elif isinstance(obj, DataFrameWithSchema):
        result = obj.copy(deep=False)
        if hasattr(obj, "json_schema"):
            result.json_schema = obj.json_schema
    elif isinstance(obj, DataFrame):
        result = DataFrameWithSchema(obj)
    elif is_list_tensor(obj):
        obj = np.array(obj)
        result = obj.view(NDArrayWithSchema)
    elif isinstance(obj, (DataFrameGroupBy, SeriesGroupBy)):
        result = obj
    elif spark_installed and isinstance(obj, SparkGroupedData):
        result = obj
    else:
        raise ValueError(f"unexpected type(obj) {type(obj)}")
    setattr(result, "table_name", name)
    return result


def get_table_name(obj):
    if spark_installed and isinstance(obj, SparkDataFrame):
        # Python class DataFrame with field self._jdf: https://github.com/apache/spark/blob/master/python/pyspark/sql/dataframe.py
        # Scala type DataFrame: https://github.com/apache/spark/blob/master/sql/core/src/main/scala/org/apache/spark/sql/package.scala
        # Scala class DataSet with field queryExecution: https://github.com/apache/spark/blob/master/sql/core/src/main/scala/org/apache/spark/sql/Dataset.scala
        # Scala fields turn into Java nullary methods
        # Py4J exposes Java methods as Python methods
        # Scala class QueryExecution with field analyzed: LogicalPlan: https://github.com/apache/spark/blob/master/sql/core/src/main/scala/org/apache/spark/sql/execution/QueryExecution.scala
        spark_query = obj._jdf.queryExecution().analyzed()  # type: ignore
        try:
            # calling spark_df.explain("extended") shows the analyzed contents
            # after spark_df.alias("foo"), analyzed contents should be SubqueryAlias
            # Scala class SuqueryAlias with field identifier: https://github.com/apache/spark/blob/master/sql/catalyst/src/main/scala/org/apache/spark/sql/catalyst/plans/logical/basicLogicalOperators.scala
            # str(..) converts the Java string into a Python string
            result = str(spark_query.identifier())
        except py4j.protocol.Py4JError:
            result = None
        return result
    if isinstance(
        obj,
        (
            NDArrayWithSchema,
            SeriesWithSchema,
            DataFrameWithSchema,
            DataFrameGroupBy,
            SeriesGroupBy,
        ),
    ) or (spark_installed and isinstance(obj, SparkGroupedData)):
        return getattr(obj, "table_name", None)
    return None


def get_index_name(obj):
    result = None
    if spark_installed and isinstance(obj, SparkDataFrameWithIndex):
        result = obj.index_name
    elif isinstance(
        obj,
        (
            SeriesWithSchema,
            DataFrameWithSchema,
            DataFrameGroupBy,
            SeriesGroupBy,
        ),
    ):
        result = obj.index.name
    return result


def get_index_names(obj):
    result = None
    if spark_installed and isinstance(obj, SparkDataFrameWithIndex):
        result = obj.index_names
    elif isinstance(
        obj,
        (
            SeriesWithSchema,
            DataFrameWithSchema,
            DataFrameGroupBy,
            SeriesGroupBy,
        ),
    ):
        result = obj.index.names
    return result


def forward_metadata(old, new):
    new = add_table_name(new, get_table_name(old))
    if isinstance(old, SparkDataFrameWithIndex):
        new = SparkDataFrameWithIndex(new, index_names=get_index_names(old))
    return new


def strip_schema(obj):
    if isinstance(obj, NDArrayWithSchema):
        result = np.array(obj)
        assert type(result) == ndarray  # pylint:disable=unidiomatic-typecheck
    elif isinstance(obj, SeriesWithSchema):
        result = Series(obj)
        assert type(result) == Series  # pylint:disable=unidiomatic-typecheck
    elif isinstance(obj, DataFrameWithSchema):
        result = DataFrame(obj)
        assert type(result) == DataFrame  # pylint:disable=unidiomatic-typecheck
    else:
        result = obj
    return result


def _dtype_to_schema(typ) -> JSON_TYPE:
    result: JSON_TYPE
    if typ is bool or issubdtype(typ, np.bool_):
        result = {"type": "boolean"}
    elif issubdtype(typ, np.unsignedinteger):
        result = {"type": "integer", "minimum": 0}
    elif issubdtype(typ, np.integer):
        result = {"type": "integer"}
    elif issubdtype(typ, np.number):
        result = {"type": "number"}
    elif issubdtype(typ, np.string_) or issubdtype(typ, np.unicode_):
        result = {"type": "string"}
    elif isinstance(typ, np.dtype):
        if typ.fields:
            props = {k: _dtype_to_schema(t) for k, t in typ.fields.items()}
            result = {"type": "object", "properties": props}
        elif typ.shape:
            result = _shape_and_dtype_to_schema(typ.shape, typ.subdtype)
        elif issubdtype(typ, np.object_):
            result = {"type": "string"}
        else:
            assert False, f"unexpected dtype {typ}"
    else:
        assert False, f"unexpected non-dtype {typ}"
    return result


def dtype_to_schema(typ) -> JSON_TYPE:
    result = _dtype_to_schema(typ)
    lale.type_checking.validate_is_schema(result)
    return result


def _shape_and_dtype_to_schema(shape, dtype) -> JSON_TYPE:
    result = _dtype_to_schema(dtype)
    for dim in reversed(shape):
        result = {"type": "array", "minItems": dim, "maxItems": dim, "items": result}
    return result


def shape_and_dtype_to_schema(shape, dtype) -> JSON_TYPE:
    result = _shape_and_dtype_to_schema(shape, dtype)
    lale.type_checking.validate_is_schema(result)
    return result


def list_tensor_to_shape_and_dtype(ls) -> Optional[Tuple[Tuple[int, ...], Type]]:
    if isinstance(ls, (int, float, str)):
        return ((), type(ls))
    if isinstance(ls, list):
        sub_result: Any = "Any"
        for item in ls:
            item_result = list_tensor_to_shape_and_dtype(item)
            if item_result is None:
                return None
            if sub_result == "Any":
                sub_result = item_result
            elif sub_result != item_result:
                return None
        if sub_result == "Any" and len(ls) == 0:
            return ((len(ls),) + (), int)
        sub_shape, sub_dtype = sub_result
        return ((len(ls),) + sub_shape, sub_dtype)
    return None


def is_list_tensor(obj) -> bool:
    if isinstance(obj, list):
        shape_and_dtype = list_tensor_to_shape_and_dtype(obj)
        return shape_and_dtype is not None
    return False


def _list_tensor_to_schema(ls) -> Optional[JSON_TYPE]:
    shape_and_dtype = list_tensor_to_shape_and_dtype(ls)
    if shape_and_dtype is None:
        return None
    result = _shape_and_dtype_to_schema(*shape_and_dtype)
    return result


def list_tensor_to_schema(ls) -> Optional[JSON_TYPE]:
    result = _list_tensor_to_schema(ls)
    if result is None:
        return None
    lale.type_checking.validate_is_schema(result)
    return result


def _ndarray_to_schema(array) -> JSON_TYPE:
    assert isinstance(array, ndarray)
    if (
        isinstance(array, NDArrayWithSchema)
        and hasattr(array, "json_schema")
        and array.json_schema is not None
    ):
        return array.json_schema
    return _shape_and_dtype_to_schema(array.shape, array.dtype)


def ndarray_to_schema(array) -> JSON_TYPE:
    result = _ndarray_to_schema(array)
    lale.type_checking.validate_is_schema(result)
    return result


def _csr_matrix_to_schema(matrix) -> JSON_TYPE:
    assert isinstance(matrix, csr_matrix)
    result = _shape_and_dtype_to_schema(matrix.shape, matrix.dtype)
    result["isSparse"] = {}  # true schema
    return result


def csr_matrix_to_schema(matrix) -> JSON_TYPE:
    result = _csr_matrix_to_schema(matrix)
    lale.type_checking.validate_is_schema(result)
    return result


def _dataframe_to_schema(df) -> JSON_TYPE:
    assert isinstance(df, DataFrame)
    if (
        isinstance(df, DataFrameWithSchema)
        and hasattr(df, "json_schema")
        and df.json_schema is not None
    ):
        return df.json_schema
    n_rows, n_columns = df.shape
    df_dtypes = df.dtypes
    assert n_columns == len(df.columns) and n_columns == len(df_dtypes)
    items = [
        {"description": str(col), **_dtype_to_schema(df_dtypes[col])}
        for col in df.columns
    ]
    result = {
        "type": "array",
        "minItems": n_rows,
        "maxItems": n_rows,
        "items": {
            "type": "array",
            "minItems": n_columns,
            "maxItems": n_columns,
            "items": items,
        },
    }
    return result


def dataframe_to_schema(df) -> JSON_TYPE:
    result = _dataframe_to_schema(df)
    lale.type_checking.validate_is_schema(result)
    return result


def _series_to_schema(series) -> JSON_TYPE:
    assert isinstance(series, Series)
    if (
        isinstance(series, SeriesWithSchema)
        and hasattr(series, "json_schema")
        and series.json_schema is not None
    ):
        return series.json_schema
    (n_rows,) = series.shape
    result = {
        "type": "array",
        "minItems": n_rows,
        "maxItems": n_rows,
        "items": {"description": str(series.name), **_dtype_to_schema(series.dtype)},
    }
    return result


def series_to_schema(series) -> JSON_TYPE:
    result = _series_to_schema(series)
    lale.type_checking.validate_is_schema(result)
    return result


def _torch_tensor_to_schema(tensor) -> JSON_TYPE:
    assert torch_installed, """Your Python environment does not have torch installed. You can install it with
    pip install torch
or with
    pip install 'lale[full]'"""
    assert isinstance(tensor, Tensor)
    result: JSON_TYPE
    # https://pytorch.org/docs/stable/tensor_attributes.html#torch-dtype
    if tensor.dtype == torch.bool:
        result = {"type": "boolean"}
    elif tensor.dtype == torch.uint8:
        result = {"type": "integer", "minimum": 0, "maximum": 255}
    elif torch.is_floating_point(tensor):
        result = {"type": "number"}
    else:
        result = {"type": "integer"}
    for dim in reversed(tensor.shape):
        result = {"type": "array", "minItems": dim, "maxItems": dim, "items": result}
    return result


def torch_tensor_to_schema(tensor) -> JSON_TYPE:
    result = _torch_tensor_to_schema(tensor)
    lale.type_checking.validate_is_schema(result)
    return result


def is_liac_arff(obj) -> bool:
    expected_types = {
        "description": str,
        "relation": str,
        "attributes": list,
        "data": list,
    }
    if not isinstance(obj, dict):
        return False

    for k, t in expected_types.items():
        if k not in obj or not isinstance(obj[k], t):
            return False
    return True


def _liac_arff_to_schema(larff) -> JSON_TYPE:
    assert is_liac_arff(
        larff
    ), """Your Python environment might contain an 'arff' package different from 'liac-arff'. You can install it with
    pip install 'liac-arff>=2.4.0'
or with
    pip install 'lale[full]'"""
    n_rows, n_columns = len(larff["data"]), len(larff["attributes"])

    def larff_type_to_schema(larff_type) -> JSON_TYPE:
        if isinstance(larff_type, str):
            a2j = {
                "numeric": "number",
                "real": "number",
                "integer": "integer",
                "string": "string",
            }
            return {"type": a2j[larff_type.lower()]}
        assert isinstance(larff_type, list)
        return {"enum": [*larff_type]}

    items = [
        {"description": attr[0], **larff_type_to_schema(attr[1])}
        for attr in larff["attributes"]
    ]
    result = {
        "type": "array",
        "minItems": n_rows,
        "maxItems": n_rows,
        "items": {
            "type": "array",
            "minItems": n_columns,
            "maxItems": n_columns,
            "items": items,
        },
    }
    return result


def liac_arff_to_schema(larff) -> JSON_TYPE:
    result = _liac_arff_to_schema(larff)
    lale.type_checking.validate_is_schema(result)
    return result


def make_optional_schema(schema: JSON_TYPE) -> JSON_TYPE:
    return {"anyOf": [schema, {"enum": [None]}]}


def _spark_df_to_schema(df) -> JSON_TYPE:
    assert spark_installed, """Your Python environment does not have spark installed. You can install it with
        pip install pyspark
"""
    assert isinstance(df, SparkDataFrameWithIndex)

    import pyspark.sql.types as stypes
    from pyspark.sql.types import StructField, StructType

    def maybe_make_optional(schema: JSON_TYPE, is_option: bool) -> JSON_TYPE:
        if is_option:
            return make_optional_schema(schema)
        return schema

    def spark_datatype_to_json_schema(dtype: stypes.DataType) -> JSON_TYPE:
        if isinstance(dtype, stypes.ArrayType):
            return {
                "type": "array",
                "items": maybe_make_optional(
                    spark_datatype_to_json_schema(dtype.elementType), dtype.containsNull
                ),
            }
        if isinstance(dtype, stypes.BooleanType):
            return {"type": "boolean"}
        if isinstance(dtype, stypes.DoubleType):
            return {"type": "number"}
        if isinstance(dtype, stypes.FloatType):
            return {"type": "number"}
        if isinstance(dtype, stypes.IntegerType):
            return {"type": "integer"}
        if isinstance(dtype, stypes.LongType):
            return {"type": "integer"}
        if isinstance(dtype, stypes.ShortType):
            return {"type": "integer"}
        if isinstance(dtype, stypes.NullType):
            return {"enum": [None]}
        if isinstance(dtype, stypes.StringType):
            return {"type": "string"}

        return {}

    def spark_struct_field_to_json_schema(f: StructField) -> JSON_TYPE:
        type_schema = spark_datatype_to_json_schema(f.dataType)
        result = maybe_make_optional(type_schema, f.nullable)

        if f.name is not None:
            result["description"] = f.name
        return result

    def spark_struct_to_json_schema(
        s: StructType, index_names, table_name: Optional[str] = None
    ) -> JSON_TYPE:
        items = [
            spark_struct_field_to_json_schema(f) for f in s if f.name not in index_names
        ]
        num_items = len(items)
        result = {
            "type": "array",
            "items": {
                "type": "array",
                "description": "rows",
                "minItems": num_items,
                "maxItems": num_items,
                "items": items,
            },
        }

        if table_name is not None:
            result["description"] = table_name

        return result

    return spark_struct_to_json_schema(df.schema, df.index_names, get_table_name(df))


def spark_df_to_schema(df) -> JSON_TYPE:
    result = _spark_df_to_schema(df)
    lale.type_checking.validate_is_schema(result)
    return result


def _to_schema(obj) -> JSON_TYPE:
    result = None
    if obj is None:
        result = {"enum": [None]}
    elif isinstance(obj, ndarray):
        result = _ndarray_to_schema(obj)
    elif isinstance(obj, csr_matrix):
        result = _csr_matrix_to_schema(obj)
    elif isinstance(obj, DataFrame):
        result = _dataframe_to_schema(obj)
    elif isinstance(obj, Series):
        result = _series_to_schema(obj)
    elif torch_installed and isinstance(obj, Tensor):
        result = _torch_tensor_to_schema(obj)
    elif is_liac_arff(obj):
        result = _liac_arff_to_schema(obj)
    elif isinstance(obj, list):
        result = _list_tensor_to_schema(obj)
    elif _is_spark_df(obj):
        result = _spark_df_to_schema(obj)
    elif lale.type_checking.is_schema(obj):
        result = obj
        # Does not need to validate again the schema
        return result  # type: ignore
    if result is None:
        raise ValueError(f"to_schema(obj), type {type(obj)}, value {obj}")
    return result


def to_schema(obj) -> JSON_TYPE:
    result = _to_schema(obj)
    lale.type_checking.validate_is_schema(result)
    return result



# Copyright 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys
from typing import Any, Tuple, overload

import numpy as np
import pandas as pd
from sklearn.utils import Bunch

from lale.datasets.data_schemas import add_table_name, get_table_name

if sys.version_info >= (3, 8):
    from typing import Literal  # raises a mypy error for <3.8
else:
    from typing_extensions import Literal

try:
    from pyspark.sql import SparkSession

    from lale.datasets.data_schemas import (  # pylint:disable=ungrouped-imports
        SparkDataFrameWithIndex,
    )

    spark_installed = True
except ImportError:
    spark_installed = False


def pandas2spark(pandas_df):
    assert spark_installed
    spark_session = (
        SparkSession.builder.master("local[2]")
        .config("spark.driver.memory", "64g")
        .getOrCreate()
    )
    name = get_table_name(pandas_df)
    if isinstance(pandas_df, pd.Series):
        pandas_df = pandas_df.to_frame()
    index_names = pandas_df.index.names
    if len(index_names) == 1 and index_names[0] is None:
        index_names = ["index"]
    cols = list(pandas_df.columns) + list(index_names)
    pandas_df = pandas_df.reset_index().reindex(columns=cols)
    spark_dataframe = spark_session.createDataFrame(pandas_df)
    spark_dataframe_with_index = SparkDataFrameWithIndex(spark_dataframe, index_names)
    return add_table_name(spark_dataframe_with_index, name)


@overload
def load_boston(return_X_y: Literal[True]) -> Tuple[Any, Any]:
    ...


@overload
def load_boston(return_X_y: Literal[False] = False) -> Bunch:
    ...


def load_boston(return_X_y: bool = False):
    data_url = "http://lib.stat.cmu.edu/datasets/boston"
    raw_df = pd.read_csv(data_url, sep=r"\s+", skiprows=22, header=None)
    data = np.hstack([raw_df.values[::2, :], raw_df.values[1::2, :2]])
    target = raw_df.values[1::2, 2]

    if return_X_y:
        return (data, target)
    else:
        return Bunch(data=data, target=target)



# Copyright 2019-2023 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import imblearn.over_sampling
import numpy as np

import lale.docstrings
import lale.operators

from ._common_schemas import (
    _hparam_n_jobs,
    _hparam_n_neighbors,
    _hparam_operator,
    _hparam_random_state,
    _hparam_sampling_strategy_anyof_neoc,
    _input_fit_schema,
    _input_predict_schema,
    _input_transform_schema,
    _output_decision_function_schema,
    _output_predict_proba_schema,
    _output_predict_schema,
    _output_transform_schema,
)
from .base_resampler import _BaseResamplerImpl


class _ADASYNImpl(_BaseResamplerImpl):
    def __init__(self, operator=None, **hyperparams):
        if operator is None:
            raise ValueError("Operator is a required argument.")

        resampler_instance = imblearn.over_sampling.ADASYN(**hyperparams)
        super().__init__(operator=operator, resampler=resampler_instance)

    def fit(self, X, y=None):
        resampler = self.resampler
        assert resampler is not None
        X, y = resampler.fit_resample(X, np.array(y))
        op = self.operator
        assert op is not None
        self.trained_operator = op.fit(X, y)
        return self


_hyperparams_schema = {
    "allOf": [
        {
            "type": "object",
            "relevantToOptimizer": ["operator"],
            "additionalProperties": False,
            "properties": {
                "operator": _hparam_operator,
                "sampling_strategy": _hparam_sampling_strategy_anyof_neoc,
                "random_state": _hparam_random_state,
                "n_neighbors": {**_hparam_n_neighbors, "default": 5},
                "n_jobs": _hparam_n_jobs,
            },
        }
    ]
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Perform over-sampling using Adaptive Synthetic (ADASYN) sampling approach for imbalanced datasets.""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.imblearn.adasyn.html",
    "import_from": "imblearn.over_sampling",
    "type": "object",
    "tags": {
        "pre": [],
        "op": [
            "transformer",
            "estimator",
            "resampler",
        ],  # transformer and estimator both as a higher-order operator
        "post": [],
    },
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_predict_proba": _input_predict_schema,
        "output_predict_proba": _output_predict_proba_schema,
        "input_decision_function": _input_predict_schema,
        "output_decision_function": _output_decision_function_schema,
    },
}


ADASYN = lale.operators.make_operator(_ADASYNImpl, _combined_schemas)

lale.docstrings.set_docstrings(ADASYN)



# Copyright 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import typing
from typing import Any, Tuple

import numpy as np

import lale.docstrings
import lale.helpers
import lale.operators
from lale.expressions import collect_set, it, replace
from lale.helpers import _ensure_pandas
from lale.lib.dataframe import count, get_columns
from lale.lib.sklearn import one_hot_encoder

from .aggregate import Aggregate
from .map import Map
from .monoid import Monoid, MonoidableOperator


class _OneHotEncoderMonoid(Monoid):
    def __init__(self, *, n_samples_seen_, feature_names_in_, categories_):
        self.n_samples_seen_ = n_samples_seen_
        self.feature_names_in_ = feature_names_in_
        self.categories_ = categories_

    def combine(self, other: "_OneHotEncoderMonoid"):
        n_samples_seen_ = self.n_samples_seen_ + other.n_samples_seen_
        assert list(self.feature_names_in_) == list(other.feature_names_in_)
        assert len(self.categories_) == len(other.categories_)
        combined_categories = [
            np.sort(
                np.unique(np.concatenate([self.categories_[i], other.categories_[i]]))
            )
            for i in range(len(self.categories_))
        ]
        return _OneHotEncoderMonoid(
            n_samples_seen_=n_samples_seen_,
            feature_names_in_=self.feature_names_in_,
            categories_=combined_categories,
        )


class _OneHotEncoderImpl(MonoidableOperator[_OneHotEncoderMonoid]):
    def __init__(
        self,
        *,
        categories="auto",
        drop=None,
        sparse=False,
        dtype="float64",
        handle_unknown="ignore",
    ):
        self._hyperparams = {
            "categories": categories,
            "drop": drop,
            "sparse": sparse,
            "dtype": dtype,
            "handle_unknown": handle_unknown,
        }

    def transform(self, X):
        if self._transformer is None:
            self._transformer = self._build_transformer()
        return self._transformer.transform(X)

    @property
    def n_samples_seen_(self):
        return getattr(self._monoid, "n_samples_seen_", 0)

    @property
    def categories_(self):
        return getattr(self._monoid, "categories_", None)

    @property
    def feature_names_in_(self):
        return getattr(self._monoid, "feature_names_in_", None)

    def from_monoid(self, monoid: _OneHotEncoderMonoid):
        self._monoid = monoid
        self.n_features_in_ = len(monoid.feature_names_in_)
        self._transformer = None

    def _build_transformer(self):
        assert self._monoid is not None
        result = Map(
            columns={
                f"{col_name}_{cat_value}": replace(
                    it[col_name],
                    {cat_value: 1},
                    handle_unknown="use_encoded_value",
                    unknown_value=0,
                )
                for col_idx, col_name in enumerate(self._monoid.feature_names_in_)
                for cat_value in self._monoid.categories_[col_idx]
            }
        )
        return result

    def to_monoid(self, batch: Tuple[Any, Any]):
        X, _ = batch
        n_samples_seen_ = count(X)
        feature_names_in_ = get_columns(X)
        agg_op = Aggregate(columns={c: collect_set(it[c]) for c in feature_names_in_})
        agg_data = agg_op.transform(X)
        agg_data = _ensure_pandas(agg_data)
        categories_ = [np.sort(agg_data.loc[0, c]) for c in feature_names_in_]
        return _OneHotEncoderMonoid(
            n_samples_seen_=n_samples_seen_,
            feature_names_in_=feature_names_in_,
            categories_=categories_,
        )


_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Relational algebra reimplementation of scikit-learn's `OneHotEncoder`_ transformer that encodes categorical features as numbers.
Works on both pandas and Spark dataframes by using `Aggregate`_ for `fit` and `Map`_ for `transform`, which in turn use the appropriate backend.

.. _`OneHotEncoder`: https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.OneHotEncoder.html
.. _`Aggregate`: https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.aggregate.html
.. _`Map`: https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.map.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.one_hot_encoder.html",
    "type": "object",
    "tags": {
        "pre": ["categoricals"],
        "op": ["transformer", "interpretable"],
        "post": [],
    },
    "properties": {
        "hyperparams": one_hot_encoder._hyperparams_schema,
        "input_fit": one_hot_encoder._input_fit_schema,
        "input_transform": one_hot_encoder._input_transform_schema,
        "output_transform": one_hot_encoder._output_transform_schema,
    },
}

OneHotEncoder = lale.operators.make_operator(_OneHotEncoderImpl, _combined_schemas)

OneHotEncoder = typing.cast(
    lale.operators.PlannedIndividualOp,
    OneHotEncoder.customize_schema(
        drop={
            "enum": [None],
            "description": "This implementation only supports `drop=None`.",
            "default": None,
        },
        sparse={
            "enum": [False],
            "description": "This implementation only supports `sparse=False`.",
            "default": False,
        },
        dtype={
            "enum": ["float64"],
            "description": "This implementation only supports `dtype='float64'`.",
            "default": "float64",
        },
        handle_unknown={
            "enum": ["ignore"],
            "description": "This implementation only supports `handle_unknown='ignore'`.",
            "default": "ignore",
        },
    ),
)

lale.docstrings.set_docstrings(OneHotEncoder)
